"""Generic Data (primarily XML & JSON) Validation Methods."""
import logging
import csv
import io
from airflow import AirflowException
from lxml import etree, isoschematron
from tulflow import process


def filter_s3_schematron(**kwargs):
    """Wrapper function for using S3 Retrieval, Schematron Filtering, and S3 Writer."""
    source_prefix = kwargs.get("source_prefix")
    dest_prefix = kwargs.get("destination_prefix")
    report_prefix = kwargs.get("report_prefix")
    bucket = kwargs.get("bucket")
    schematron_file = kwargs.get("schematron_filename")
    access_id = kwargs.get("access_id")
    access_secret = kwargs.get("access_secret")
    if kwargs.get("dag"):
        run_id = kwargs.get("dag").dag_id
    else:
        run_id = "no-dag-provided"
    if kwargs.get("timestamp"):
        timestamp = kwargs.get("timestamp")
    else:
        timestamp = "no-timestamp-provided"

    # create invalid records reporting csv
    csv_in_mem = io.StringIO()
    invalid_csv = csv.DictWriter(csv_in_mem, fieldnames=["id", "report", "record", "source_file"])
    invalid_csv.writeheader()

    # get schematron doc & return lxml.etree.Schematron validator
    schematron_doc = process.get_github_content("tulibraries/aggregator_mdx", schematron_file)
    schematron = isoschematron.Schematron(etree.fromstring(schematron_doc), store_report=True)
    total_filter_count = 0
    total_record_count = 0
    for s3_key in process.list_s3_content(bucket, access_id, access_secret, source_prefix):
        logging.info("Validating & Filtering File: %s", s3_key)
        s3_content = process.get_s3_content(bucket, s3_key, access_id, access_secret)
        s3_xml = etree.fromstring(s3_content)
        invalid_xml = etree.Element("collection")
        invalid_xml.attrib["dag-id"] = run_id
        invalid_xml.attrib["dag-timestamp"] = timestamp
        filter_count = 0
        record_count = 0
        for record in s3_xml.iterchildren():
            record_count += 1
            total_record_count += 1
            if not schematron.validate(record):
                record_id = record.get("airflow-record-id")
                logging.error("Invalid record found: %s", record_id)
                s3_xml.remove(record)
                filter_count += 1
                invalid_csv.writerow({
                    "id": record_id,
                    "report": schematron_failed_validation_text(schematron.validation_report),
                    "record": identifier_or_full_record(record),
                    "source_file": f"https://s3.console.aws.amazon.com/s3/object/{bucket}/{s3_key}"
                })
        total_filter_count += filter_count
        filename = s3_key.replace(source_prefix, dest_prefix)
        updated_s3_xml = etree.tostring(s3_xml, encoding="utf-8").decode("utf-8")
        process.generate_s3_object(updated_s3_xml, bucket, filename, access_id, access_secret)
        if filter_count == record_count and record_count != 0:
            logging.warning(f"All records filtered from {filename}. record_count: {record_count}")

    invalid_filename = report_prefix + "-invalid.csv"
    logging.info("Total Filter Count: %s", total_filter_count)
    logging.info("Invalid Records report: https://%s.s3.amazonaws.com/%s", bucket, invalid_filename)
    process.generate_s3_object(csv_in_mem.getvalue(), bucket, invalid_filename, access_id, access_secret)
    if total_filter_count == total_record_count and total_record_count != 0:
        raise AirflowException(f"All records were filtered out: {total_record_count}")
    return {"filtered": total_filter_count}

def report_s3_schematron(**kwargs):
    """Wrapper function for using S3 Retrieval, Schematron Reporting, and S3 Writer."""
    source_prefix = kwargs.get("source_prefix")
    dest_prefix = kwargs.get("destination_prefix")
    bucket = kwargs.get("bucket")
    schematron_file = kwargs.get("schematron_filename")
    access_id = kwargs.get("access_id")
    access_secret = kwargs.get("access_secret")

    # create reporting csv
    csv_in_mem = io.StringIO()
    report_csv = csv.DictWriter(csv_in_mem, fieldnames=["id", "report", "record", "source_file"])
    report_csv.writeheader()

    # get schematron doc & return lxml.etree.Schematron validator
    schematron_doc = process.get_github_content("tulibraries/aggregator_mdx", schematron_file)
    schematron = isoschematron.Schematron(etree.fromstring(schematron_doc), store_report=True)

    # Iterate through S3 Files, Validate, & Save Report to CSV
    total_transform_count = 0
    for s3_key in process.list_s3_content(bucket, access_id, access_secret, source_prefix):
        logging.info("Validating & Reporting On File: %s", s3_key)
        s3_content = process.get_s3_content(bucket, s3_key, access_id, access_secret)
        s3_xml = etree.fromstring(s3_content)
        for record in s3_xml.iterchildren():
            total_transform_count += 1
            record_id = record.get("airflow-record-id")
            logging.info("Ran report on record: %s", record_id)
            schematron.validate(record)
            report_csv.writerow({
                "id": record_id,
                "report": schematron_failed_validation_text(schematron.validation_report),
                "record": identifier_or_full_record(record),
                "source_file": f"https://s3.console.aws.amazon.com/s3/object/{bucket}/{s3_key}"
            })
    report_filename = dest_prefix + "-report.csv"
    logging.info("Records report: https://%s.s3.amazonaws.com/%s", bucket, report_filename)
    logging.info("Total Transform Count: %s", total_transform_count)
    process.generate_s3_object(csv_in_mem.getvalue(), bucket, report_filename, access_id, access_secret)

    return {"transformed": total_transform_count}

def identifier_or_full_record(record, identifier_xpath="./dcterms:identifier/text()", identifier_namespaces={"dcterms": "http://purl.org/dc/terms/"}):
    indentifiers = record.xpath(identifier_xpath, namespaces=identifier_namespaces)
    if indentifiers:
        return "\n".join(indentifiers)
    else:
        return etree.tostring(record, encoding="utf-8").decode("utf-8")


def schematron_failed_validation_text(validation_report):
    return "\n".join(
        validation_report.xpath("./svrl:failed-assert/svrl:text/text()",namespaces={"svrl": "http://purl.oclc.org/dsdl/svrl"})
                    )



"""
tulflow.harvest
~~~~~~~~~~~~~~~
This module contains objects to harvest data from one given location to another.
"""
import hashlib
import io
import logging
import pandas
import sickle
from lxml import etree
from sickle import Sickle
from sickle.models import xml_to_dict
from sickle.oaiexceptions import NoRecordsMatch
from tulflow import process

NS = {
    "marc21": "http://www.loc.gov/MARC21/slim",
    "oai": "http://www.openarchives.org/OAI/2.0/"
    }


def oai_to_s3(**kwargs):
    """Wrapper function for using OAI Harvest, Default Processor, and S3 Writer."""
    kwargs["harvest_params"] = {
        "metadataPrefix": kwargs.get("metadata_prefix"),
        "from": kwargs.get("harvest_from_date"),
        "until": kwargs.get("harvest_until_date")
    }
    dag_id = kwargs["dag"].dag_id
    dag_start_date = kwargs["timestamp"]

    oai_sets = generate_oai_sets(**kwargs)
    all_processed = []
    sets_with_no_records = []
    if oai_sets:
        for oai_set in oai_sets:
            kwargs["harvest_params"]["set"] = oai_set
            data = harvest_oai(**kwargs)
            if data == []:
                sets_with_no_records.append(oai_set)
                logging.info("Skipping processing % set because it has no data.", oai_set)
                continue
            outdir = dag_s3_prefix(dag_id, dag_start_date)
            processed = process_xml(data, dag_write_string_to_s3, outdir, **kwargs)
            all_processed.append(processed)
    else:
        data = harvest_oai(**kwargs)
        if data == []:
            sets_with_no_records.append(oai_set)
        outdir = dag_s3_prefix(dag_id, dag_start_date)
        processed = process_xml(data, dag_write_string_to_s3, outdir, **kwargs)
        all_processed.append(processed)
    all_updated = sum([set['updated'] for set in all_processed])
    all_deleted = sum([set['deleted'] for set in all_processed])
    logging.info("Total OAI Records Harvested & Processed: %s", all_updated)
    logging.info("Total OAI Records Harvest & Marked for Deletion: %s", all_deleted)
    logging.info("Total sets with no records: %s", len(sets_with_no_records))
    logging.info("Sets with no records %s", sets_with_no_records)
    return {"updated": all_updated, "deleted": all_deleted, "sets_with_no_records": sets_with_no_records}


def generate_oai_sets(**kwargs):
    """Generate the oai sets we want to harvest."""
    all_sets = bool(kwargs.get("all_sets"))
    included_sets = kwargs.get("included_sets")
    excluded_sets = kwargs.get("excluded_sets")
    oai_endpoint = kwargs.get("oai_endpoint")

    if all_sets:
        logging.info("Seeing All Sets Needed.")
        return []
    elif included_sets:
        logging.info("Seeing SetSpec List.")
        if not isinstance(included_sets, list):
            return [included_sets]
        return included_sets
    elif excluded_sets:
        logging.info("Seeing Excluded SetSpec List.")
        if not isinstance(excluded_sets, list):
            excluded_sets = [excluded_sets]
        list_sets = Sickle(oai_endpoint).ListSets()
        all_sets = [oai_set.xml.find("oai:setSpec", namespaces=NS).text for oai_set in list_sets]
        remaining_sets = list(set(all_sets) - set(excluded_sets))
        logging.info(remaining_sets)
        return remaining_sets
    return []


class HarvestIterator(sickle.iterator.OAIItemIterator):
    def next(self):
        """Return the next record/header/set."""
        while True:
            for item in self._items:
                mapped = self.mapper(item)
                if self.ignore_deleted and mapped.deleted:
                    continue
                if hasattr(mapped, 'metadata') and mapped.metadata == None:
                    logging.info("Skipping record with no metadata: %s", mapped.header.identifier)
                    continue
                return mapped
            if self.resumption_token and self.resumption_token.token:
                self._next_response()
            else:
                raise StopIteration
    pass

# TODO: Remove if https://github.com/mloesch/sickle/pull/47 gets merged.
class HarvestRecord(sickle.models.Record):
    def get_metadata(self):
        # We want to get record/metadata/<container>/*
        # <container> would be the element ``dc``
        # in the ``oai_dc`` case.
        meta_data = self.xml.find('.//' + self._oai_namespace + 'metadata')
        if meta_data != None:
            return xml_to_dict(meta_data.getchildren()[0], strip_ns=self._strip_ns)
    pass

def harvest_oai(**kwargs):
    """Create OAI ListRecords Iterator for Harvesting Data."""
    oai_endpoint = kwargs.get("oai_endpoint")
    harvest_params = kwargs.get("harvest_params")
    logging.info("Harvesting from %s", oai_endpoint)
    logging.info("Harvesting %s", harvest_params)
    sickle = Sickle(oai_endpoint, retry_status_codes=[500,503,504], max_retries=3)

    class_mapping = harvest_params.get("class_mapping", {
        "ListRecords": HarvestRecord,
        })
    iterator = harvest_params.get("iterator", HarvestIterator)
    for key in class_mapping:
        sickle.class_mapping[key] = class_mapping[key]

    sickle.iterator = iterator

    try:
        return sickle.ListRecords(**harvest_params)
    except NoRecordsMatch:
        logging.info("No records found.")
        return []


class OaiXml:
    """oai-pmh xml etree wrapper"""
    def __init__(self, dag_id, timestamp):
        etree.register_namespace("oai", "http://www.openarchives.org/OAI/2.0/")
        etree.register_namespace("marc21", "http://www.loc.gov/MARC21/slim")
        self.root = etree.Element("{http://www.openarchives.org/OAI/2.0/}collection")
        self.root.attrib["dag-id"] = dag_id
        self.root.attrib["dag-timestamp"] = timestamp

    def append(self, record):
        self.root.append(record)

    def tostring(self):
        return etree.tostring(self.root, encoding="utf-8").decode("utf-8")


def process_xml(data, writer, outdir, **kwargs):
    """Process & Write XML data to S3."""
    parser = kwargs.get("parser")
    records_per_file = kwargs.get("records_per_file")
    if kwargs.get("dag"):
        run_id = kwargs.get("dag").dag_id
    else:
        run_id = "no-dag-provided"
    if kwargs.get("timestamp"):
        timestamp = kwargs.get("timestamp")
    else:
        timestamp = "no-timestamp-provided"
    if not records_per_file:
        records_per_file = 1000

    count = deleted_count = 0
    oai_updates = OaiXml(run_id, timestamp)
    oai_deletes = OaiXml(run_id, timestamp)
    logging.info("Processing XML")

    for record in data:
        record_id = record.header.identifier
        record = record.xml
        record.attrib["airflow-record-id"] = record_id
        if parser:
            record = parser(record, **kwargs)
        if record.xpath(".//oai:header[@status='deleted']", namespaces=NS):
            logging.info("Added record %s to deleted xml file(s)", record_id)
            deleted_count += 1
            oai_deletes.append(record)

            if deleted_count % int(records_per_file) == 0:
                writer(oai_deletes.tostring(), outdir + "/deleted", **kwargs)
                oai_deletes = OaiXml(run_id, timestamp)
        else:
            logging.info("Added record %s to new-updated xml file", record_id)
            count += 1
            oai_updates.append(record)
            if count % int(records_per_file) == 0:
                writer(oai_updates.tostring(), outdir + "/new-updated", **kwargs)
                oai_updates = OaiXml(run_id, timestamp)
    writer(oai_updates.tostring(), outdir + "/new-updated", **kwargs)
    writer(oai_deletes.tostring(), outdir + "/deleted", **kwargs)
    logging.info("OAI Records Harvested & Processed: %s", count)
    logging.info("OAI Records Harvest & Marked for Deletion: %s", deleted_count)
    return {"updated": count, "deleted": deleted_count}


def perform_xml_lookup_with_cache():
    cache = {}

    def perform_xml_lookup(oai_record, **kwargs):
        """Parse additions/updates & add boundwiths."""

        if len(cache) == 0:
            logging.info("*** Fetching CSV lookup file from s3 ***")
            access_id = kwargs.get("access_id")
            access_secret = kwargs.get("access_secret")
            bucket = kwargs.get("bucket_name")
            lookup_key = kwargs.get("lookup_key")
            csv_data = process.get_s3_content(bucket, lookup_key, access_id, access_secret)
            cache["value"] = pandas.read_csv(io.BytesIO(csv_data), header=0)

        lookup_csv = cache["value"]

        for record in oai_record.xpath(".//marc21:record", namespaces=NS):
            record_id = process.get_record_001(record)
            logging.info("Reading in Record %s", record_id)
            parent_txt = lookup_csv.loc[lookup_csv.child_id == int(record_id), "parent_xml"].values
            if len(set(parent_txt)) >= 1:
                logging.info("Child XML record found %s", record_id)
                for parent_node in parent_txt[0].split("||"):
                    try:
                        record.append(etree.fromstring(parent_node))
                    except etree.XMLSyntaxError as error:
                        logging.error("Problem with string syntax:")
                        logging.error(error)
                        logging.error(parent_node)
        return oai_record


    return perform_xml_lookup


def dag_write_string_to_s3(string, prefix, **kwargs):
    """Push a string in memory to s3 with a defined prefix"""
    access_id = kwargs.get("access_id")
    access_secret = kwargs.get("access_secret")
    bucket_name = kwargs.get("bucket_name")
    logging.info("Writing to S3 Bucket %s", bucket_name)

    our_hash = hashlib.md5(string.encode("utf-8")).hexdigest()
    filename = "{}/{}".format(prefix, our_hash)
    process.generate_s3_object(string, bucket_name, filename, access_id, access_secret)


def write_log(string, prefix, **kwargs):
    """Write the data to logging info."""
    prefix = prefix
    logging.info(prefix)
    string = string
    logging.info(string)


def dag_s3_prefix(dag_id, timestamp):
    """Define the prefix that will be prepended to all files created by this dag run"""
    return "{}/{}".format(dag_id, timestamp)



"""Tests suite for tulflow harvest (Functions for harvesting OAI in Airflow Tasks)."""
import unittest
import boto3
import httpretty
from lxml import etree
from moto import mock_s3
from tulflow import process

class TestDataProcessInteractions(unittest.TestCase):
    """Test Class for data processing functions."""

    def test_expand_alma_sftp_tarball_pass(self):
        """Test Untarring AlmaSFTP S3 Tarball bytestream to XML."""
        test_key = "almasftp/alma_bibs__new_1.xml.tar.gz"
        test_object = open("tests/fixtures/alma_bibs__new_1.xml.tar.gz", "rb").read()
        test_run = process.expand_alma_sftp_tarball(test_key, test_object)
        self.assertEqual(type(test_run).__name__, "bytes")
        self.assertEqual(test_run, open("tests/fixtures/alma_bibs__new_1.xml", "rb").read())

    def test_expand_alma_sftp_tarball_empty(self):
        """Test Untarring AlmaSFTP S3 Tarball bytestream to XML."""
        test_key = "almasftp/alma_bibs__new_1.xml.tar.gz"
        test_object = open("tests/fixtures/alma_bibs__empty.xml.tar.gz", "rb").read()
        with self.assertLogs("tulflow_process") as log:
            test_run = process.expand_alma_sftp_tarball(test_key, test_object)
        self.assertEqual(test_run, None)
        logs = ["ERROR:tulflow_process:S3 Object is empty.", "ERROR:tulflow_process:" + test_key]
        self.assertEqual(log.output, logs)

    def test_expand_alma_sftp_tarball_multi(self):
        """Test Untarring AlmaSFTP S3 Tarball bytestream to XML."""
        test_key = "almasftp/alma_bibs__new_1.xml.tar.gz"
        test_object = open("tests/fixtures/alma_bibs__multi.xml.tar.gz", "rb").read()
        with self.assertLogs("tulflow_process") as log:
            test_run = process.expand_alma_sftp_tarball(test_key, test_object)
        self.assertEqual(test_run, None)
        logs = [
            "ERROR:tulflow_process:S3 Object has more than 1 member, which is unexpected.",
            "ERROR:tulflow_process:" + test_key
        ]
        self.assertEqual(log.output, logs)

    def test_add_marc21xml_root_namespace(self):
        """Test converting ALMASFTP XML Collection document as bytes
        to lxml.etree.Element with MARC21 as default namespace."""
        test_bytes = open("tests/fixtures/alma_bibs__new_1.xml", "rb").read()
        test_out = etree.fromstring(open("tests/fixtures/alma_bibs__new_1_ns.xml", "rb").read())
        test_run = process.add_marc21xml_root_ns(test_bytes)
        self.assertEqual(etree.tostring(test_run), etree.tostring(test_out))
        self.assertIn("{http://www.loc.gov/MARC21/slim}", test_run.tag)

    def test_add_marc21xml_root_namespace_dup(self):
        """Test converting ALMASFTP XML Collection document as bytes
        to lxml.etree.Element with MARC21 as default namespace."""
        test_bytes = open("tests/fixtures/alma_bibs__new_1_ns.xml", "rb").read()
        test_out = etree.fromstring(open("tests/fixtures/alma_bibs__new_1_ns.xml", "rb").read())
        test_run = process.add_marc21xml_root_ns(test_bytes)
        self.assertEqual(etree.tostring(test_run), etree.tostring(test_out))
        self.assertIn("{http://www.loc.gov/MARC21/slim}", test_run.tag)

    def test_get_record_001(self):
        """Test validating & returning a MARC/XML OO1 field."""
        test_xml = etree.fromstring(open("tests/fixtures/record_001.xml", "rb").read())
        test_run = process.get_record_001(test_xml)
        self.assertEqual(test_run, "991022063789703811")

    def test_get_record_001_empty(self):
        """Test validating & returning a MARC/XML OO1 field."""
        test_xml = etree.fromstring(open("tests/fixtures/record_001_empty.xml", "rb").read())
        with self.assertLogs("tulflow_process") as log:
            test_run = process.get_record_001(test_xml)
        self.assertEqual(test_run, None)
        logs = [
            "ERROR:tulflow_process:Record without an 001 MMS Identifier:",
            "ERROR:tulflow_process:" + str(etree.tostring(test_xml))
        ]
        self.assertEqual(log.output, logs)

    def test_get_record_001_missing(self):
        """Test validating & returning a MARC/XML OO1 field."""
        test_xml = etree.fromstring(open("tests/fixtures/record_001_missing.xml", "rb").read())
        with self.assertLogs("tulflow_process") as log:
            test_run = process.get_record_001(test_xml)
        self.assertEqual(test_run, None)
        logs = [
            "ERROR:tulflow_process:Record without an 001 MMS Identifier:",
            "ERROR:tulflow_process:" + str(etree.tostring(test_xml))
        ]
        self.assertEqual(log.output, logs)

    def test_get_record_001_dup(self):
        """Test validating & returning a MARC/XML OO1 field."""
        test_xml = etree.fromstring(open("tests/fixtures/record_001_dup.xml", "rb").read())
        test_run = process.get_record_001(test_xml)
        with self.assertLogs("tulflow_process") as log:
            test_run = process.get_record_001(test_xml)
        self.assertEqual(test_run, None)
        logs = [
            "ERROR:tulflow_process:Record with multiple 001 MMS Identifiers:",
            "ERROR:tulflow_process:" + str(etree.tostring(test_xml))
        ]
        self.assertEqual(log.output, logs)

    def test_generate_bw_parent_new_field(self):
        desired_xml = b"""<marc21:datafield xmlns:marc21="http://www.loc.gov/MARC21/slim" ind1=" " ind2=" " tag="ADF"><marc21:subfield code="a">FAKE_PARENT_ID</marc21:subfield></marc21:datafield>"""
        test_run = process.generate_bw_parent_field("FAKE_PARENT_ID")
        self.assertEqual(etree.tostring(test_run), desired_xml)

    @httpretty.activate
    def test_get_github_content_exists(self):
        """Test getting existing file from GitHub."""
        httpretty.register_uri(
            httpretty.GET,
            "https://raw.github.com/tulibraries/aggregator_mdx/main/transforms/temple.xsl",
            body=open("tests/fixtures/temple.xsl").read()
        )
        repository = "tulibraries/aggregator_mdx"
        filename = "transforms/temple.xsl"

        test_run = process.get_github_content(repository, filename)
        self.assertEqual(test_run, open("tests/fixtures/temple.xsl", "rb").read())

    @httpretty.activate
    def test_get_github_content_missing(self):
        """Test getting missing file from GitHub & return appropriate error."""
        httpretty.register_uri(
            httpretty.GET,
            "https://raw.github.com/tulibraries/aggregator_mdx/main/transforms/temple-fake.xsl",
            status=404
        )
        repository = "tulibraries/aggregator_mdx"
        filename = "transforms/temple-fake.xsl"

        with self.assertLogs() as log:
            with self.assertRaises(SystemExit) as pytest_wrapped_e:
                process.get_github_content(repository, filename)
        self.assertEqual(pytest_wrapped_e.exception.code, 1)
        self.assertIn("ERROR:root:404 Client Error: Not Found for url: https://raw.github.com/tulibraries/aggregator_mdx/main/transforms/temple-fake.xsl", log.output)

class TestS3ProcessInteractions(unittest.TestCase):
    """Test Class for S3 data processing functions."""

    @mock_s3
    def test_remove_s3_object(self):
        bucket = "test_bucket"
        key = "test_key"
        access_id = "test_access_id"
        access_secret = "test_access_secret"
        conn = boto3.client("s3", aws_access_key_id=access_id, aws_secret_access_key=access_secret)
        conn.create_bucket(Bucket=bucket)
        conn.put_object(Bucket=bucket, Key=key, Body="test content")
        test_content_exists = conn.get_object(Bucket=bucket, Key=key)
        test_object_exists = conn.list_objects(Bucket=bucket)
        self.assertEqual(test_content_exists["Body"].read(), b"test content")
        self.assertEqual(test_content_exists["ResponseMetadata"]["HTTPStatusCode"], 200)
        self.assertEqual(test_object_exists["Contents"][0]["Key"], key)
        process.remove_s3_object(bucket, key, access_id, access_secret)
        test_object_gone = conn.list_objects(Bucket=bucket)
        self.assertEqual(test_object_gone["ResponseMetadata"]["HTTPStatusCode"], 200)
        self.assertEqual(test_object_gone.get("Contents"), None)

    @mock_s3
    def test_get_s3_content(self):
        bucket = "test_bucket"
        key = "test_key_2"
        access_id = "test_access_id"
        access_secret = "test_access_secret"
        conn = boto3.client("s3", aws_access_key_id=access_id, aws_secret_access_key=access_secret)
        conn.create_bucket(Bucket=bucket)
        conn.put_object(Bucket=bucket, Key=key, Body="test more content")
        test_content_exists = conn.get_object(Bucket=bucket, Key=key)
        test_object_exists = conn.list_objects(Bucket=bucket)
        self.assertEqual(test_content_exists["Body"].read(), b"test more content")
        self.assertEqual(test_content_exists["ResponseMetadata"]["HTTPStatusCode"], 200)
        self.assertEqual(test_object_exists["Contents"][0]["Key"], key)
        test_run = process.get_s3_content(bucket, key, access_id, access_secret)
        self.assertEqual(test_run, b"test more content")

    @mock_s3
    def test_list_s3_content(self):
        bucket = "test_bucket"
        prefix = "test_prefix"
        access_id = "test_access_id"
        access_secret = "test_access_secret"
        conn = boto3.client("s3", aws_access_key_id=access_id, aws_secret_access_key=access_secret)
        conn.create_bucket(Bucket=bucket)
        conn.put_object(Bucket=bucket, Key="test_prefix/item1", Body="test listing items")
        conn.put_object(Bucket=bucket, Key="item2", Body="including more items")
        conn.put_object(Bucket=bucket, Key="item2.magic", Body="and even more items")
        test_run_all = process.list_s3_content(bucket, access_id, access_secret)
        self.assertEqual(test_run_all, ["item2", "item2.magic", "test_prefix/item1"])
        test_run_pref = process.list_s3_content(bucket, access_id, access_secret, prefix=prefix)
        self.assertEqual(test_run_pref, ["test_prefix/item1"])

    @mock_s3
    def test_genereate_s3_object(self):
        bucket = "test_bucket"
        key = "test_key_2"
        body = b"<test>even more content</test>"
        access_id = "test_access_id"
        access_secret = "test_access_secret"
        conn = boto3.client("s3", aws_access_key_id=access_id, aws_secret_access_key=access_secret)
        conn.create_bucket(Bucket=bucket)
        process.generate_s3_object(body, bucket, key, access_id, access_secret)
        test_content_exists = conn.get_object(Bucket=bucket, Key=key)
        test_object_exists = conn.list_objects(Bucket=bucket)
        self.assertEqual(test_content_exists["Body"].read(), body)
        self.assertEqual(test_content_exists["ResponseMetadata"]["HTTPStatusCode"], 200)
        self.assertEqual(test_object_exists["Contents"][0]["Key"], key)



"""Tests suite for tulflow tasks (generic Airflow Tasks as Functions)."""
import unittest
import pytest
from unittest.mock import patch
from airflow.contrib.operators.slack_webhook_operator import SlackWebhookOperator
from airflow.operators.http_operator import SimpleHttpOperator
from airflow.hooks.base_hook import BaseHook
from airflow.models import Connection, DAG, TaskInstance
from airflow.utils import timezone
from tulflow.tasks import create_sc_collection, get_solr_url, refresh_sc_collection_for_alias, swap_sc_alias

DEFAULT_DATE = timezone.datetime(2019, 8, 16)


class TestSolrCloudTasks(unittest.TestCase):
    """Test Class for SolrCloud Tasks."""


    @patch.object(BaseHook, "get_connection", return_value=Connection(
        conn_id='SOLRCLOUD',
        conn_type='http',
        host='http://localhost',
        password='puppies'
        )
    )
    def test_create_sc_collection(self, mocker):
        """Test create_sc_collection operator instance contains expected config values."""
        dag = DAG(dag_id='test_create_sc_collection', start_date=DEFAULT_DATE)
        operator = create_sc_collection(dag, 'SOLRCLOUD', 'test-collection', '2', 'test-configset')
        task_instance = TaskInstance(task=operator, execution_date=DEFAULT_DATE)

        self.assertEqual("SOLRCLOUD", operator.http_conn_id)
        self.assertEqual("CREATE", operator.data['action'])
        self.assertEqual("test-collection", operator.data['name'])
        self.assertEqual("1", operator.data['numShards'])
        self.assertEqual("2", operator.data['replicationFactor'])
        self.assertEqual("test-configset", operator.data['collection.configName'])
        self.assertEqual("create_collection", task_instance.task_id)


    @patch.object(BaseHook, "get_connection", return_value=Connection(
        conn_id='SOLRCLOUD',
        conn_type='http',
        host='http://localhost',
        password='puppies'
        )
    )
    def test_swap_sc_alias(self, mocker):
        """Test swap_sc_alias operator instance contains expected config values."""
        dag = DAG(dag_id='test_swap_sc_alias', start_date=DEFAULT_DATE)
        operator = swap_sc_alias(dag, 'SOLRCLOUD', 'new-collection', 'my-alias')
        task_instance = TaskInstance(task=operator, execution_date=DEFAULT_DATE)

        self.assertEqual("SOLRCLOUD", operator.http_conn_id)
        self.assertEqual("CREATEALIAS", operator.data['action'])
        self.assertEqual("my-alias", operator.data['name'])
        self.assertEqual(["new-collection"], operator.data['collections'])
        self.assertEqual("solr_alias_swap", task_instance.task_id)


    @patch.object(BaseHook, "get_connection", return_value=Connection(
        conn_id='SOLRCLOUD',
        conn_type='http',
        host='http://localhost',
        password='puppies',
        login="wholoves"
        )
    )
    def test_refresh_sc_collection_for_alias(self, mocker):
        """Test refresh_sc_collection_for_alias task instance contains expected config values."""
        dag = DAG(dag_id='test_refresh_sc_collection_for_alias', start_date=DEFAULT_DATE)
        task = refresh_sc_collection_for_alias(dag=dag, sc_conn=mocker, sc_coll_name='my-collection', sc_alias='my-alias', configset="my-configset")
        task_instance = TaskInstance(task=task, execution_date=DEFAULT_DATE)

        self.assertEqual("my-collection", task.op_kwargs['collection'])
        self.assertEqual("my-alias", task.op_kwargs['alias'])
        self.assertEqual("my-configset", task.op_kwargs['configset'])
        self.assertEqual("refresh_sc_collection_for_alias", task_instance.task_id)


class TestTasksGetSolrUrl(unittest.TestCase):
    """Tests for tasks.get_solr_url function."""

    def test_get_solr_url_without_http_in_host(self):
        conn = Connection(host="example.com", port="8983")
        core = "foo"
        self.assertEqual(get_solr_url(conn, core), "http://example.com:8983/solr/foo")

    def test_get_solr_url_with_http_in_host(self):
        conn = Connection(host="https://example.com", port="8983")
        core = "foo"
        self.assertEqual(get_solr_url(conn, core), "https://example.com:8983/solr/foo")

    def test_get_solr_url_without_port(self):
        conn = Connection(host="https://example.com")
        core = "foo"
        self.assertEqual(get_solr_url(conn, core), "https://example.com/solr/foo")



"""Tests suite for tulflow.validate (functions for validating XML or JSON in Airflow Tasks)."""
import unittest
import boto3
from lxml import etree
from unittest import mock
from moto import mock_s3
from tulflow import process, validate
from unittest.mock import patch
from airflow import AirflowException

class TestSchematronFiltering(unittest.TestCase):
    """Test Class for functions that filtering XML with Schematron."""
    maxDiff = None
    kwargs = {
        "source_prefix": "dpla_test/transformed",
        "destination_prefix": "dpla_test/transformed-filtered",
        "report_prefix": "dpla_test/harvest_filter",
        "bucket": "tulib-airflow-test",
        "schematron_filename": "validations/padigital_reqd_fields.sch",
        "access_id": "kittens",
        "access_secret": "puppies"
    }

    @mock_s3
    @patch("tulflow.process.get_github_content")
    def test_filter_s3_schematron_all_valid(self, mocked_get_github_content):
        """Test Filtering S3 XML File with Schematron."""
        # setup kwargs for test runs
        access_id = self.kwargs.get("access_id")
        access_secret = self.kwargs.get("access_secret")
        bucket = self.kwargs.get("bucket")
        test_key = self.kwargs.get("source_prefix") + "/sch-oai-valid.xml"

        # create expected mocked s3 resources
        conn = boto3.client("s3", aws_access_key_id=access_id, aws_secret_access_key=access_secret)
        conn.create_bucket(Bucket=bucket)
        conn.put_object(Bucket=bucket, Key=test_key, Body=open("tests/fixtures/sch-oai-valid.xml").read())
        test_content_exists = conn.get_object(Bucket=bucket, Key=test_key)
        test_object_exists = conn.list_objects(Bucket=bucket)
        self.assertEqual(test_content_exists["Body"].read(), open("tests/fixtures/sch-oai-valid.xml", "rb").read())
        self.assertEqual(test_content_exists["ResponseMetadata"]["HTTPStatusCode"], 200)
        self.assertEqual(test_object_exists["Contents"][0]["Key"], test_key)

        # mock schematron github retrieval response
        mocked_get_github_content.return_value = open("tests/fixtures/sch-sample.sch").read()

        # run tests
        with self.assertLogs() as log:
            response = validate.filter_s3_schematron(**self.kwargs)
        self.assertIn("INFO:root:Validating & Filtering File: dpla_test/transformed/sch-oai-valid.xml", log.output)
        test_valid_objects = conn.list_objects(Bucket=bucket, Prefix=self.kwargs.get("destination_prefix") + "/")
        test_valid_objects_ar = [object.get("Key") for object in test_valid_objects["Contents"]]
        self.assertEqual(test_valid_objects["ResponseMetadata"]["HTTPStatusCode"], 200)
        self.assertEqual(test_valid_objects_ar, ["dpla_test/transformed-filtered/sch-oai-valid.xml"])
        self.assertEqual(response, { "filtered": 0 })

        test_invalid_objects = conn.list_objects(Bucket=bucket, Prefix=self.kwargs.get("report_prefix") + "-invalid.csv")
        test_invalid_objects_ar = [object.get("Key") for object in test_invalid_objects["Contents"]]
        self.assertEqual(test_invalid_objects["ResponseMetadata"]["HTTPStatusCode"], 200)
        self.assertEqual(test_invalid_objects_ar, ["dpla_test/harvest_filter-invalid.csv"])
        test_invalid_content = conn.get_object(Bucket=bucket, Key="dpla_test/harvest_filter-invalid.csv")
        self.assertEqual(test_invalid_content["Body"].read(), b"""id,report,record,source_file\r\n""")


    @mock_s3
    @patch("tulflow.process.get_github_content")
    def test_filter_s3_schematron_all_invalid(self, mocked_get_github_content):
        """Test Filtering S3 XML File with Schematron."""
        # setup kwargs for test runs
        access_id = self.kwargs.get("access_id")
        access_secret = self.kwargs.get("access_secret")
        bucket = self.kwargs.get("bucket")
        test_key = self.kwargs.get("source_prefix") + "/sch-oai-invalid.xml"

        # create expected mocked s3 resources
        conn = boto3.client("s3", aws_access_key_id=access_id, aws_secret_access_key=access_secret)
        conn.create_bucket(Bucket=bucket)
        conn.put_object(Bucket=bucket, Key=test_key, Body=open("tests/fixtures/sch-oai-invalid.xml").read())
        test_content_exists = conn.get_object(Bucket=bucket, Key=test_key)
        test_object_exists = conn.list_objects(Bucket=bucket)
        self.assertEqual(test_content_exists["Body"].read(), open("tests/fixtures/sch-oai-invalid.xml", "rb").read())
        self.assertEqual(test_content_exists["ResponseMetadata"]["HTTPStatusCode"], 200)
        self.assertEqual(test_object_exists["Contents"][0]["Key"], test_key)

        # mock schematron github retrieval response
        mocked_get_github_content.return_value = open("tests/fixtures/sch-sample.sch").read()

        # run tests
        with self.assertLogs() as log:
            with self.assertRaises(AirflowException) as context:
                validate.filter_s3_schematron(**self.kwargs)

        errors = "All records were filtered out: 5"
        exception = AirflowException(errors)
        self.assertEqual(str(exception), str(context.exception))

        self.assertIn("INFO:root:Validating & Filtering File: dpla_test/transformed/sch-oai-invalid.xml", log.output)
        self.assertIn("ERROR:root:Invalid record found: invalid-missingtitle", log.output[1])
        self.assertIn("ERROR:root:Invalid record found: invalid-missingrights", log.output[2])
        self.assertIn('WARNING:root:All records filtered from dpla_test/transformed-filtered/sch-oai-invalid.xml. record_count: 5', log.output[6])
        self.assertEqual(len(log.output), 9)

        test_valid_objects = conn.list_objects(Bucket=bucket, Prefix=self.kwargs.get("destination_prefix") + "/")
        test_valid_objects_ar = [object.get("Key") for object in test_valid_objects["Contents"]]
        test_valid_content = conn.get_object(Bucket=bucket, Key="dpla_test/transformed-filtered/sch-oai-invalid.xml")
        self.assertEqual(test_valid_objects["ResponseMetadata"]["HTTPStatusCode"], 200)
        self.assertEqual(test_valid_objects_ar, ["dpla_test/transformed-filtered/sch-oai-invalid.xml"])
        self.assertEqual(test_valid_content["Body"].read(), b"""<metadata xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:dcterms="http://purl.org/dc/terms/" xmlns:edm="http://www.europeana.eu/schemas/edm/" xmlns:oai_dc="http://www.openarchives.org/OAI/2.0/oai_dc/" xmlns:dpla="http://dp.la/about/map/" xmlns:schema="http://schema.org" xmlns:oai="http://www.openarchives.org/OAI/2.0/" xmlns:oai_qdc="http://worldcat.org/xmlschemas/qdc-1.0/">\n   </metadata>""")

        invalid_prefix = self.kwargs.get("report_prefix") + "-invalid.csv"
        test_invalid_objects = conn.list_objects(Bucket=bucket, Prefix=invalid_prefix)
        self.assertEqual(test_invalid_objects["ResponseMetadata"]["HTTPStatusCode"], 200)
        test_invalid_objects_ar = [object.get("Key") for object in test_invalid_objects["Contents"]]
        self.assertEqual(len(test_invalid_objects_ar), 1)
        self.assertIn("dpla_test/harvest_filter-invalid.csv", test_invalid_objects_ar)
        test_invalid_content = conn.get_object(Bucket=bucket, Key="dpla_test/harvest_filter-invalid.csv")["Body"].read()
        self.assertIn(b"""id,report,record,source_file\r\n""", test_invalid_content)


    @mock_s3
    @patch("tulflow.process.get_github_content")
    def test_filter_s3_schematron_mix(self, mocked_get_github_content):
        """Test Filtering S3 XML File with Schematron."""
        # setup kwargs for test runs
        access_id = self.kwargs.get("access_id")
        access_secret = self.kwargs.get("access_secret")
        bucket = self.kwargs.get("bucket")
        test_key = self.kwargs.get("source_prefix") + "/sch-oai-mix.xml"

        # create expected mocked s3 resources
        conn = boto3.client("s3", aws_access_key_id=access_id, aws_secret_access_key=access_secret)
        conn.create_bucket(Bucket=bucket)
        conn.put_object(Bucket=bucket, Key=test_key, Body=open("tests/fixtures/sch-oai-mix.xml").read())
        test_content_exists = conn.get_object(Bucket=bucket, Key=test_key)
        test_object_exists = conn.list_objects(Bucket=bucket)
        self.assertEqual(test_content_exists["Body"].read(), open("tests/fixtures/sch-oai-mix.xml", "rb").read())
        self.assertEqual(test_content_exists["ResponseMetadata"]["HTTPStatusCode"], 200)
        self.assertEqual(test_object_exists["Contents"][0]["Key"], test_key)

        # mock schematron github retrieval response
        mocked_get_github_content.return_value = open("tests/fixtures/sch-sample.sch").read()

        # run tests
        with self.assertLogs() as log:
            response = validate.filter_s3_schematron(**self.kwargs)
        self.assertIn("INFO:root:Validating & Filtering File: dpla_test/transformed/sch-oai-mix.xml", log.output)
        self.assertIn("ERROR:root:Invalid record found: invalid-missingtitle", log.output[1])
        self.assertIn("ERROR:root:Invalid record found: invalid-missingrights", log.output[2])
        self.assertEqual(len(log.output), 8)
        self.assertEqual(response, { "filtered": 5 })

        test_valid_objects = conn.list_objects(Bucket=bucket, Prefix=self.kwargs.get("destination_prefix") + "/")
        test_valid_objects_ar = [object.get("Key") for object in test_valid_objects["Contents"]]
        test_valid_content = conn.get_object(Bucket=bucket, Key="dpla_test/transformed-filtered/sch-oai-mix.xml")["Body"].read()
        self.assertEqual(test_valid_objects["ResponseMetadata"]["HTTPStatusCode"], 200)
        self.assertEqual(test_valid_objects_ar, ["dpla_test/transformed-filtered/sch-oai-mix.xml"])
        self.assertIn(b'<oai_dc:dc airflow-record-id="valid">', test_valid_content)
        self.assertIn(b"<dcterms:identifier>valid</dcterms:identifier>", test_valid_content)
        self.assertIn(b"<dcterms:identifier>valid2</dcterms:identifier>", test_valid_content)
        self.assertIn(b"<dcterms:identifier>valid3</dcterms:identifier>", test_valid_content)

        invalid_prefix = self.kwargs.get("report_prefix") + "-invalid.csv"
        test_invalid_objects = conn.list_objects(Bucket=bucket, Prefix=invalid_prefix)
        self.assertEqual(test_invalid_objects["ResponseMetadata"]["HTTPStatusCode"], 200)
        test_invalid_objects_ar = [object.get("Key") for object in test_invalid_objects["Contents"]]
        self.assertIn("dpla_test/harvest_filter-invalid.csv", test_invalid_objects_ar)
        self.assertEqual(len(test_invalid_objects_ar), 1)
        test_invalid_content = conn.get_object(Bucket=bucket, Key="dpla_test/harvest_filter-invalid.csv")["Body"].read()
        self.assertIn(b"invalid-missingtitle", test_invalid_content)
        self.assertIn(b"here must be a rights statement", test_invalid_content)


    @mock_s3
    @patch("tulflow.process.get_github_content")
    def test_filter_s3_schematron_empty(self, mocked_get_github_content):
        """Test Filtering S3 XML File with Schematron."""
        # setup kwargs for test runs
        access_id = self.kwargs.get("access_id")
        access_secret = self.kwargs.get("access_secret")
        bucket = self.kwargs.get("bucket")
        test_key = self.kwargs.get("source_prefix") + "/sch-oai-empty.xml"

        # create expected mocked s3 resources
        conn = boto3.client("s3", aws_access_key_id=access_id, aws_secret_access_key=access_secret)
        conn.create_bucket(Bucket=bucket)
        conn.put_object(Bucket=bucket, Key=test_key, Body=open("tests/fixtures/sch-oai-empty.xml").read())
        test_content_exists = conn.get_object(Bucket=bucket, Key=test_key)
        test_object_exists = conn.list_objects(Bucket=bucket)
        self.assertEqual(test_content_exists["Body"].read(), open("tests/fixtures/sch-oai-empty.xml", "rb").read())
        self.assertEqual(test_content_exists["ResponseMetadata"]["HTTPStatusCode"], 200)
        self.assertEqual(test_object_exists["Contents"][0]["Key"], test_key)

        # mock schematron github retrieval response
        mocked_get_github_content.return_value = open("tests/fixtures/sch-sample.sch").read()

        # run tests
        with self.assertLogs() as log:
            response = validate.filter_s3_schematron(**self.kwargs)
        self.assertIn("INFO:root:Validating & Filtering File: dpla_test/transformed/sch-oai-empty.xml", log.output)
        self.assertIn("INFO:root:Invalid Records report: https://tulib-airflow-test.s3.amazonaws.com/dpla_test/harvest_filter-invalid.csv", log.output)
        test_objects = conn.list_objects(Bucket=bucket, Prefix=self.kwargs.get("destination_prefix") + "/")
        test_objects_ar = [object.get("Key") for object in test_objects["Contents"]]
        test_content = conn.get_object(Bucket=bucket, Key="dpla_test/transformed-filtered/sch-oai-empty.xml")
        self.assertEqual(test_objects["ResponseMetadata"]["HTTPStatusCode"], 200)
        self.assertEqual(test_objects_ar, ["dpla_test/transformed-filtered/sch-oai-empty.xml"])
        self.assertEqual(test_content["Body"].read(), b"""<metadata xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:dcterms="http://purl.org/dc/terms/" xmlns:edm="http://www.europeana.eu/schemas/edm/" xmlns:oai_dc="http://www.openarchives.org/OAI/2.0/oai_dc/" xmlns:dpla="http://dp.la/about/map/" xmlns:schema="http://schema.org" xmlns:oai="http://www.openarchives.org/OAI/2.0/" xmlns:oai_qdc="http://worldcat.org/xmlschemas/qdc-1.0/">\n</metadata>""")
        self.assertEqual(response, { "filtered": 0 })

    @mock_s3
    @patch("tulflow.process.get_github_content")
    def test_filter_s3_schematron_none(self, mocked_get_github_content):
        """Test Filtering S3 XML File with Schematron."""
        # setup kwargs for test runs
        access_id = self.kwargs.get("access_id")
        access_secret = self.kwargs.get("access_secret")
        bucket = self.kwargs.get("bucket")
        test_key = self.kwargs.get("source_prefix") + "/sch-oai-none.xml"

        # create expected mocked s3 resources
        conn = boto3.client("s3", aws_access_key_id=access_id, aws_secret_access_key=access_secret)
        conn.create_bucket(Bucket=bucket)

        # mock schematron github retrieval response
        mocked_get_github_content.return_value = open("tests/fixtures/sch-sample.sch").read()

        # run tests
        response = validate.filter_s3_schematron(**self.kwargs)
        test_objects = conn.list_objects(Bucket=bucket, Prefix=self.kwargs.get("destination_prefix") + "/")
        self.assertEqual(test_objects["ResponseMetadata"]["HTTPStatusCode"], 200)
        self.assertEqual(test_objects.get("Contents"), None)
        self.assertEqual(response, { "filtered": 0 })

class TestSchematronReporting(unittest.TestCase):
    """Test Class for functions that generate reports on XML validated with Schematron."""
    maxDiff = None
    kwargs = {
        "source_prefix": "dpla_test/transformed",
        "destination_prefix": "dpla_test/transformed-filtered",
        "bucket": "tulib-airflow-test",
        "schematron_filename": "validations/padigital_reqd_fields.sch",
        "access_id": "kittens",
        "access_secret": "puppies"
    }

    @mock_s3
    @patch("tulflow.process.get_github_content")
    def test_report_s3_schematron_all_valid(self, mocked_get_github_content):
        """Test Reporting on Valid S3 XML File Validated with Schematron."""
        # setup kwargs for test runs
        access_id = self.kwargs.get("access_id")
        access_secret = self.kwargs.get("access_secret")
        bucket = self.kwargs.get("bucket")
        test_key = self.kwargs.get("source_prefix") + "/sch-oai-valid.xml"

        # create expected mocked s3 resources
        conn = boto3.client("s3", aws_access_key_id=access_id, aws_secret_access_key=access_secret)
        conn.create_bucket(Bucket=bucket)
        conn.put_object(Bucket=bucket, Key=test_key, Body=open("tests/fixtures/sch-oai-valid.xml").read())
        test_content_exists = conn.get_object(Bucket=bucket, Key=test_key)
        test_object_exists = conn.list_objects(Bucket=bucket)
        self.assertEqual(test_content_exists["Body"].read(), open("tests/fixtures/sch-oai-valid.xml", "rb").read())
        self.assertEqual(test_content_exists["ResponseMetadata"]["HTTPStatusCode"], 200)
        self.assertEqual(test_object_exists["Contents"][0]["Key"], test_key)

        # mock schematron github retrieval response
        mocked_get_github_content.return_value = open("tests/fixtures/sch-sample.sch").read()

        # run tests
        with self.assertLogs() as log:
            response = validate.report_s3_schematron(**self.kwargs)
        self.assertIn("INFO:root:Validating & Reporting On File: dpla_test/transformed/sch-oai-valid.xml", log.output)
        test_report_objects = conn.list_objects(Bucket=bucket, Prefix=self.kwargs.get("destination_prefix") + "-report.csv")
        self.assertEqual(test_report_objects["ResponseMetadata"]["HTTPStatusCode"], 200)
        test_report_objects_ar = [object.get("Key") for object in test_report_objects["Contents"]]
        self.assertEqual(test_report_objects_ar, ["dpla_test/transformed-filtered-report.csv"])
        test_report_content = conn.get_object(Bucket=bucket, Key="dpla_test/transformed-filtered-report.csv")["Body"].read()
        self.assertEqual(test_report_content, b'id,report,record,source_file\r\nvalid,,valid,https://s3.console.aws.amazon.com/s3/object/tulib-airflow-test/dpla_test/transformed/sch-oai-valid.xml\r\nvalid2,,valid2,https://s3.console.aws.amazon.com/s3/object/tulib-airflow-test/dpla_test/transformed/sch-oai-valid.xml\r\nvalid3,,valid3,https://s3.console.aws.amazon.com/s3/object/tulib-airflow-test/dpla_test/transformed/sch-oai-valid.xml\r\n')
        self.assertIn(b'id,report,record,source_file\r\nvalid', test_report_content)
        self.assertEqual(response, { "transformed": 3 })


    @mock_s3
    @patch("tulflow.process.get_github_content")
    def test_report_s3_schematron_all_invalid(self, mocked_get_github_content):
        """Test Reporting on Invalid S3 XML File Validated with Schematron."""
        # setup kwargs for test runs
        access_id = self.kwargs.get("access_id")
        access_secret = self.kwargs.get("access_secret")
        bucket = self.kwargs.get("bucket")
        test_key = self.kwargs.get("source_prefix") + "/sch-oai-invalid.xml"

        # create expected mocked s3 resources
        conn = boto3.client("s3", aws_access_key_id=access_id, aws_secret_access_key=access_secret)
        conn.create_bucket(Bucket=bucket)
        conn.put_object(Bucket=bucket, Key=test_key, Body=open("tests/fixtures/sch-oai-invalid.xml").read())
        test_content_exists = conn.get_object(Bucket=bucket, Key=test_key)
        test_object_exists = conn.list_objects(Bucket=bucket)
        self.assertEqual(test_content_exists["Body"].read(), open("tests/fixtures/sch-oai-invalid.xml", "rb").read())
        self.assertEqual(test_content_exists["ResponseMetadata"]["HTTPStatusCode"], 200)
        self.assertEqual(test_object_exists["Contents"][0]["Key"], test_key)

        # mock schematron github retrieval response
        mocked_get_github_content.return_value = open("tests/fixtures/sch-sample.sch").read()

        # run tests
        with self.assertLogs() as log:
            response = validate.report_s3_schematron(**self.kwargs)
        self.assertIn("INFO:root:Validating & Reporting On File: dpla_test/transformed/sch-oai-invalid.xml", log.output)
        test_report_objects = conn.list_objects(Bucket=bucket, Prefix=self.kwargs.get("destination_prefix") + "-report.csv")
        self.assertEqual(test_report_objects["ResponseMetadata"]["HTTPStatusCode"], 200)
        test_report_objects_ar = [object.get("Key") for object in test_report_objects["Contents"]]
        self.assertEqual(test_report_objects_ar, ["dpla_test/transformed-filtered-report.csv"])
        test_report_content = conn.get_object(Bucket=bucket, Key="dpla_test/transformed-filtered-report.csv")["Body"].read()
        self.assertIn(b'id,report,record,source_file\r\ninvalid-missingtitle', test_report_content)
        self.assertIn(b'\r\ninvalid-missingrights', test_report_content)
        self.assertIn(b'\r\ninvalid-missingitemurl', test_report_content)
        self.assertEqual(response, { "transformed": 5 })


    @mock_s3
    @patch("tulflow.process.get_github_content")
    def test_report_s3_schematron_mix(self, mocked_get_github_content):
        """Test Reporting on Valid & Invalid S3 XML File Validated with Schematron."""
        # setup kwargs for test runs
        access_id = self.kwargs.get("access_id")
        access_secret = self.kwargs.get("access_secret")
        bucket = self.kwargs.get("bucket")
        test_key = self.kwargs.get("source_prefix") + "/sch-oai-mix.xml"

        # create expected mocked s3 resources
        conn = boto3.client("s3", aws_access_key_id=access_id, aws_secret_access_key=access_secret)
        conn.create_bucket(Bucket=bucket)
        conn.put_object(Bucket=bucket, Key=test_key, Body=open("tests/fixtures/sch-oai-mix.xml").read())
        test_content_exists = conn.get_object(Bucket=bucket, Key=test_key)
        test_object_exists = conn.list_objects(Bucket=bucket)
        self.assertEqual(test_content_exists["Body"].read(), open("tests/fixtures/sch-oai-mix.xml", "rb").read())
        self.assertEqual(test_content_exists["ResponseMetadata"]["HTTPStatusCode"], 200)
        self.assertEqual(test_object_exists["Contents"][0]["Key"], test_key)

        # mock schematron github retrieval response
        mocked_get_github_content.return_value = open("tests/fixtures/sch-sample.sch").read()

        # run tests
        with self.assertLogs() as log:
            response = validate.report_s3_schematron(**self.kwargs)
        self.assertIn("INFO:root:Validating & Reporting On File: dpla_test/transformed/sch-oai-mix.xml", log.output)
        test_report_objects = conn.list_objects(Bucket=bucket, Prefix=self.kwargs.get("destination_prefix") + "-report.csv")
        self.assertEqual(test_report_objects["ResponseMetadata"]["HTTPStatusCode"], 200)
        test_report_objects_ar = [object.get("Key") for object in test_report_objects["Contents"]]
        self.assertEqual(test_report_objects_ar, ["dpla_test/transformed-filtered-report.csv"])
        test_report_content = conn.get_object(Bucket=bucket, Key="dpla_test/transformed-filtered-report.csv")["Body"].read()
        self.assertIn(b'id,report,record,source_file\r\nvalid', test_report_content)
        self.assertIn(b'\r\ninvalid-missingtitle', test_report_content)
        self.assertIn(b'\r\ninvalid-missingrights,', test_report_content)
        self.assertIn(b'\r\ninvalid-missingitemurl,', test_report_content)
        self.assertEqual(response, { "transformed": 8 })

    @mock_s3
    @patch("tulflow.process.get_github_content")
    def test_report_s3_schematron_empty(self, mocked_get_github_content):
        """Test Reporting on Empty S3 XML File Validated with Schematron."""
        # setup kwargs for test runs
        access_id = self.kwargs.get("access_id")
        access_secret = self.kwargs.get("access_secret")
        bucket = self.kwargs.get("bucket")
        test_key = self.kwargs.get("source_prefix") + "/sch-oai-empty.xml"

        # create expected mocked s3 resources
        conn = boto3.client("s3", aws_access_key_id=access_id, aws_secret_access_key=access_secret)
        conn.create_bucket(Bucket=bucket)
        conn.put_object(Bucket=bucket, Key=test_key, Body=open("tests/fixtures/sch-oai-empty.xml").read())
        test_content_exists = conn.get_object(Bucket=bucket, Key=test_key)
        test_object_exists = conn.list_objects(Bucket=bucket)
        self.assertEqual(test_content_exists["Body"].read(), open("tests/fixtures/sch-oai-empty.xml", "rb").read())
        self.assertEqual(test_content_exists["ResponseMetadata"]["HTTPStatusCode"], 200)
        self.assertEqual(test_object_exists["Contents"][0]["Key"], test_key)

        # mock schematron github retrieval response
        mocked_get_github_content.return_value = open("tests/fixtures/sch-sample.sch").read()

        # run tests
        with self.assertLogs() as log:
            response = validate.report_s3_schematron(**self.kwargs)
        self.assertIn("INFO:root:Validating & Reporting On File: dpla_test/transformed/sch-oai-empty.xml", log.output)
        test_report_objects = conn.list_objects(Bucket=bucket, Prefix=self.kwargs.get("destination_prefix") + "-report.csv")
        self.assertEqual(test_report_objects["ResponseMetadata"]["HTTPStatusCode"], 200)
        test_report_objects_ar = [object.get("Key") for object in test_report_objects["Contents"]]
        self.assertEqual(test_report_objects_ar, ["dpla_test/transformed-filtered-report.csv"])
        test_report_content = conn.get_object(Bucket=bucket, Key="dpla_test/transformed-filtered-report.csv")["Body"].read()
        self.assertIn(b'id,report,record,source_file\r\n', test_report_content)
        self.assertEqual(response, { "transformed": 0 })

    @mock_s3
    @patch("tulflow.process.get_github_content")
    def test_report_s3_schematron_none(self, mocked_get_github_content):
        """Test Reporting on No provided S3 XML File Validated with Schematron."""
        # setup kwargs for test runs
        access_id = self.kwargs.get("access_id")
        access_secret = self.kwargs.get("access_secret")
        bucket = self.kwargs.get("bucket")

        # create expected mocked s3 resources
        conn = boto3.client("s3", aws_access_key_id=access_id, aws_secret_access_key=access_secret)
        conn.create_bucket(Bucket=bucket)

        # mock schematron github retrieval response
        mocked_get_github_content.return_value = open("tests/fixtures/sch-sample.sch").read()

        # run tests
        with self.assertLogs() as log:
            response = validate.report_s3_schematron(**self.kwargs)
        self.assertEqual(log.output, ["INFO:root:Records report: https://tulib-airflow-test.s3.amazonaws.com/dpla_test/transformed-filtered-report.csv", 'INFO:root:Total Transform Count: 0'])
        test_report_objects = conn.list_objects(Bucket=bucket, Prefix=self.kwargs.get("destination_prefix") + "-report.csv")
        self.assertEqual(test_report_objects["ResponseMetadata"]["HTTPStatusCode"], 200)
        test_report_objects_ar = [object.get("Key") for object in test_report_objects["Contents"]]
        self.assertEqual(test_report_objects_ar, ["dpla_test/transformed-filtered-report.csv"])
        test_report_content = conn.get_object(Bucket=bucket, Key="dpla_test/transformed-filtered-report.csv")["Body"].read()
        self.assertEqual(b'id,report,record,source_file\r\n', test_report_content)
        self.assertEqual(response, { "transformed": 0 })

    def test_schematron_failed_validation_text(self):
        single_failure = etree.fromstring(b'<svrl:schematron-output xmlns:svrl="http://purl.oclc.org/dsdl/svrl" xmlns:xs="http://www.w3.org/2001/XMLSchema" xmlns:schold="http://www.ascc.net/xml/schematron" xmlns:sch="http://www.ascc.net/xml/schematron" xmlns:iso="http://purl.oclc.org/dsdl/schematron" xmlns:dcterms="http://purl.org/dc/terms/" xmlns:edm="http://www.europeana.eu/schemas/edm/" xmlns:oai_dc="http://www.openarchives.org/OAI/2.0/oai_dc/" title="" schemaVersion=""><!--  &#160;\n\t\t  &#160;\n\t\t  &#160;\n\t\t --><svrl:ns-prefix-in-attribute-values uri="http://purl.org/dc/terms/" prefix="dcterms"/><svrl:ns-prefix-in-attribute-values uri="http://www.europeana.eu/schemas/edm/" prefix="edm"/><svrl:ns-prefix-in-attribute-values uri="http://www.openarchives.org/OAI/2.0/oai_dc/" prefix="oai_dc"/><svrl:active-pattern id="RequiredElementsPattern" name="Required PA Digital Elements"/><svrl:fired-rule context="oai_dc:dc"/><svrl:failed-assert test="edm:preview" id="Required1" role="error" location="/*[local-name()=\'dc\' and namespace-uri()=\'http://www.openarchives.org/OAI/2.0/oai_dc/\']"><svrl:text>There must be a thumbnail URL</svrl:text></svrl:failed-assert><svrl:active-pattern id="ThumbnailURLElementPattern" name="Additional Thumbnail URL Requirements"/></svrl:schematron-output>')
        response = validate.schematron_failed_validation_text(single_failure)
        self.assertEqual(response, 'There must be a thumbnail URL')

        multiple_failures = etree.fromstring(b'<svrl:schematron-output xmlns:svrl="http://purl.oclc.org/dsdl/svrl" xmlns:xs="http://www.w3.org/2001/XMLSchema" xmlns:schold="http://www.ascc.net/xml/schematron" xmlns:sch="http://www.ascc.net/xml/schematron" xmlns:iso="http://purl.oclc.org/dsdl/schematron" xmlns:dcterms="http://purl.org/dc/terms/" xmlns:edm="http://www.europeana.eu/schemas/edm/" xmlns:oai_dc="http://www.openarchives.org/OAI/2.0/oai_dc/" title="" schemaVersion=""><!--  &#160;\n\t\t  &#160;\n\t\t  &#160;\n\t\t --><svrl:ns-prefix-in-attribute-values uri="http://purl.org/dc/terms/" prefix="dcterms"/><svrl:ns-prefix-in-attribute-values uri="http://www.europeana.eu/schemas/edm/" prefix="edm"/><svrl:ns-prefix-in-attribute-values uri="http://www.openarchives.org/OAI/2.0/oai_dc/" prefix="oai_dc"/> <svrl:active-pattern id="RequiredElementsPattern" name="Required PA Digital Elements"/><svrl:fired-rule context="oai_dc:dc"/><svrl:failed-assert test="edm:preview" id="Required1" role="error" location="/*[local-name()=\'dc\' and namespace-uri()=\'http://www.openarchives.org/OAI/2.0/oai_dc/\']"><svrl:text>Needs Another Thing</svrl:text></svrl:failed-assert><svrl:active-pattern id="OtherPattern" name="Other Thing"/> <svrl:active-pattern id="RequiredElementsPattern" name="Required PA Digital Elements"/><svrl:fired-rule context="oai_dc:dc"/><svrl:failed-assert test="edm:preview" id="Required1" role="error" location="/*[local-name()=\'dc\' and namespace-uri()=\'http://www.openarchives.org/OAI/2.0/oai_dc/\']"><svrl:text>There must be a thumbnail URL</svrl:text></svrl:failed-assert><svrl:active-pattern id="ThumbnailURLElementPattern" name="Additional Thumbnail URL Requirements"/></svrl:schematron-output>')
        response = validate.schematron_failed_validation_text(multiple_failures)
        self.assertEqual(response, 'Needs Another Thing\nThere must be a thumbnail URL')

        no_failures = etree.fromstring(b'<svrl:schematron-output xmlns:svrl="http://purl.oclc.org/dsdl/svrl" xmlns:xs="http://www.w3.org/2001/XMLSchema" xmlns:schold="http://www.ascc.net/xml/schematron" xmlns:sch="http://www.ascc.net/xml/schematron" xmlns:iso="http://purl.oclc.org/dsdl/schematron" xmlns:dcterms="http://purl.org/dc/terms/" xmlns:edm="http://www.europeana.eu/schemas/edm/" xmlns:oai_dc="http://www.openarchives.org/OAI/2.0/oai_dc/" title="" schemaVersion=""><!--  &#160;\n\t\t  &#160;\n\t\t  &#160;\n\t\t --><svrl:ns-prefix-in-attribute-values uri="http://purl.org/dc/terms/" prefix="dcterms"/><svrl:ns-prefix-in-attribute-values uri="http://www.europeana.eu/schemas/edm/" prefix="edm"/><svrl:ns-prefix-in-attribute-values uri="http://www.openarchives.org/OAI/2.0/oai_dc/" prefix="oai_dc"/><svrl:active-pattern id="RequiredElementsPattern" name="Required PA Digital Elements"/><svrl:fired-rule context="oai_dc:dc"/><svrl:active-pattern id="ThumbnailURLElementPattern" name="Additional Thumbnail URL Requirements"/><svrl:fired-rule context="oai_dc:dc/edm:preview"/></svrl:schematron-output>')
        response = validate.schematron_failed_validation_text(no_failures)
        self.assertEqual(response, '')

    def test_identifier_or_full_record(self):
        record = etree.fromstring(b'<oai_dc:dc xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:dcterms="http://purl.org/dc/terms/" xmlns:dpla="http://dp.la/about/map/" xmlns:edm="http://www.europeana.eu/schemas/edm/" xmlns:oai="http://www.openarchives.org/OAI/2.0/" xmlns:oai_dc="http://www.openarchives.org/OAI/2.0/oai_dc/" xmlns:oai_qdc="http://worldcat.org/xmlschemas/qdc-1.0/" xmlns:oclc="http://purl.org/oclc/terms/" xmlns:oclcdc="http://worldcat.org/xmlschemas/oclcdc-1.0/" xmlns:oclcterms="http://purl.org/oclc/terms/" xmlns:schema="http://schema.org" xmlns:padig="http://padigital.org/ns" xmlns:svcs="http://rdfs.org/sioc/services" airflow-record-id="pitt:00add0682m">\n   <dcterms:isPartOf>Historic Pittsburgh Book Collection</dcterms:isPartOf>\n   <dcterms:title>Thomas Mellon and his times</dcterms:title>\n   <dcterms:creator>Mellon, Thomas , 1813-1908</dcterms:creator>\n   <dcterms:subject>Mellon family</dcterms:subject>\n   <dcterms:description>pt. I. Family history.--pt. II. Autobiography.</dcterms:description>\n   <dcterms:description>Printed for his family and descendants exclusively.</dcterms:description>\n   <dcterms:publisher>W. G. Johnston &amp; Co., printers</dcterms:publisher>\n   <edm:dataProvider>University of Pittsburgh</edm:dataProvider>\n   <dcterms:date>1885</dcterms:date>\n   <dcterms:type>Text</dcterms:type>\n   <dcterms:format>bibliography</dcterms:format>\n   <dcterms:format>biography</dcterms:format>\n    <dcterms:identifier>pitt:00add0682m</dcterms:identifier>\n   <dcterms:language>eng</dcterms:language>\n   <dcterms:rights>No Copyright - United States. The organization that has made the Item available believes that the Item is in the Public Domain under the laws of the United States, but a determination was not made as to its copyright status under the copyright laws of other countries. The Item may not be in the Public Domain under the laws of other countries. Please refer to the organization that has made the Item available for more information.</dcterms:rights>\n   <edm:rights>http://rightsstatements.org/vocab/NoC-US/1.0/</edm:rights>\n   <edm:preview>http://historicpittsburgh.org/islandora/object/pitt%3A00add0682m/datastream/TN/view/Thomas%20Mellon%20and%20his%20times.jpg</edm:preview>\n   <edm:isShownAt>http://historicpittsburgh.org/islandora/object/pitt%3A00add0682m</edm:isShownAt>\n   <dpla:intermediateProvider>Historic Pittsburgh</dpla:intermediateProvider>\n   <edm:provider>PA Digital</edm:provider>\n</oai_dc:dc>\n')
        identifiers = validate.identifier_or_full_record(record)
        self.assertEqual(identifiers, "pitt:00add0682m")

        multiple_identifiers = etree.fromstring('<oai_dc:dc xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:dcterms="http://purl.org/dc/terms/" xmlns:dpla="http://dp.la/about/map/" xmlns:edm="http://www.europeana.eu/schemas/edm/" xmlns:oai="http://www.openarchives.org/OAI/2.0/" xmlns:oai_dc="http://www.openarchives.org/OAI/2.0/oai_dc/" xmlns:oai_qdc="http://worldcat.org/xmlschemas/qdc-1.0/" xmlns:oclc="http://purl.org/oclc/terms/" xmlns:oclcdc="http://worldcat.org/xmlschemas/oclcdc-1.0/" xmlns:oclcterms="http://purl.org/oclc/terms/" xmlns:schema="http://schema.org" xmlns:padig="http://padigital.org/ns" xmlns:svcs="http://rdfs.org/sioc/services" airflow-record-id="pitt:00add0682m">\n   <dcterms:isPartOf>Historic Pittsburgh Book Collection</dcterms:isPartOf>\n   <dcterms:title>Thomas Mellon and his times</dcterms:title>\n   <dcterms:creator>Mellon, Thomas , 1813-1908</dcterms:creator>\n   <dcterms:subject>Mellon family</dcterms:subject>\n   <dcterms:description>pt. I. Family history.--pt. II. Autobiography.</dcterms:description>\n   <dcterms:description>Printed for his family and descendants exclusively.</dcterms:description>\n   <dcterms:publisher>W. G. Johnston &amp; Co., printers</dcterms:publisher>\n   <edm:dataProvider>University of Pittsburgh</edm:dataProvider>\n   <dcterms:date>1885</dcterms:date>\n   <dcterms:type>Text</dcterms:type>\n   <dcterms:format>bibliography</dcterms:format>\n   <dcterms:format>biography</dcterms:format>\n  <dcterms:identifier>pitt:00add0682m</dcterms:identifier>\n   <dcterms:language>eng</dcterms:language>\n   <dcterms:identifier>other:identifier</dcterms:identifier>\n   <dcterms:rights>No Copyright - United States. The organization that has made the Item available believes that the Item is in the Public Domain under the laws of the United States, but a determination was not made as to its copyright status under the copyright laws of other countries. The Item may not be in the Public Domain under the laws of other countries. Please refer to the organization that has made the Item available for more information.</dcterms:rights>\n   <edm:rights>http://rightsstatements.org/vocab/NoC-US/1.0/</edm:rights>\n   <edm:preview>http://historicpittsburgh.org/islandora/object/pitt%3A00add0682m/datastream/TN/view/Thomas%20Mellon%20and%20his%20times.jpg</edm:preview>\n   <edm:isShownAt>http://historicpittsburgh.org/islandora/object/pitt%3A00add0682m</edm:isShownAt>\n   <dpla:intermediateProvider>Historic Pittsburgh</dpla:intermediateProvider>\n   <edm:provider>PA Digital</edm:provider>\n</oai_dc:dc>\n')
        identifiers = validate.identifier_or_full_record(multiple_identifiers)
        self.assertEqual(identifiers, "pitt:00add0682m\nother:identifier")

        no_identifiers_bytestring = '<oai_dc:dc xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:dcterms="http://purl.org/dc/terms/" xmlns:dpla="http://dp.la/about/map/" xmlns:edm="http://www.europeana.eu/schemas/edm/" xmlns:oai="http://www.openarchives.org/OAI/2.0/" xmlns:oai_dc="http://www.openarchives.org/OAI/2.0/oai_dc/" xmlns:oai_qdc="http://worldcat.org/xmlschemas/qdc-1.0/" xmlns:oclc="http://purl.org/oclc/terms/" xmlns:oclcdc="http://worldcat.org/xmlschemas/oclcdc-1.0/" xmlns:oclcterms="http://purl.org/oclc/terms/" xmlns:schema="http://schema.org" xmlns:padig="http://padigital.org/ns" xmlns:svcs="http://rdfs.org/sioc/services" airflow-record-id="pitt:00add0682m">\n   <dcterms:isPartOf>Historic Pittsburgh Book Collection</dcterms:isPartOf>\n   <dcterms:title>Thomas Mellon and his times</dcterms:title>\n   <dcterms:creator>Mellon, Thomas , 1813-1908</dcterms:creator>\n   <dcterms:subject>Mellon family</dcterms:subject>\n   <dcterms:description>pt. I. Family history.--pt. II. Autobiography.</dcterms:description>\n   <dcterms:description>Printed for his family and descendants exclusively.</dcterms:description>\n   <dcterms:publisher>W. G. Johnston &amp; Co., printers</dcterms:publisher>\n   <edm:dataProvider>University of Pittsburgh</edm:dataProvider>\n   <dcterms:date>1885</dcterms:date>\n   <dcterms:type>Text</dcterms:type>\n   <dcterms:format>bibliography</dcterms:format>\n   <dcterms:format>biography</dcterms:format>\n   <dcterms:language>eng</dcterms:language>\n   <dcterms:rights>No Copyright - United States. The organization that has made the Item available believes that the Item is in the Public Domain under the laws of the United States, but a determination was not made as to its copyright status under the copyright laws of other countries. The Item may not be in the Public Domain under the laws of other countries. Please refer to the organization that has made the Item available for more information.</dcterms:rights>\n   <edm:rights>http://rightsstatements.org/vocab/NoC-US/1.0/</edm:rights>\n   <edm:preview>http://historicpittsburgh.org/islandora/object/pitt%3A00add0682m/datastream/TN/view/Thomas%20Mellon%20and%20his%20times.jpg</edm:preview>\n   <edm:isShownAt>http://historicpittsburgh.org/islandora/object/pitt%3A00add0682m</edm:isShownAt>\n   <dpla:intermediateProvider>Historic Pittsburgh</dpla:intermediateProvider>\n   <edm:provider>PA Digital</edm:provider>\n</oai_dc:dc>'
        no_identifiers = etree.fromstring(no_identifiers_bytestring)
        identifiers = validate.identifier_or_full_record(no_identifiers)
        self.assertEqual(identifiers, no_identifiers_bytestring)



"""
tulflow.transform
~~~~~~~~~~~~~~~
This module contains objects to transform data using a known transform language.
"""
import hashlib
import logging
import os
import subprocess
import sys
import requests
from lxml import etree
from tulflow import process
from pathlib import Path

# pylint: disable=unexpected-keyword-arg
def transform_s3_xsl(**kwargs):
    """Transform & Write XML data to S3 using Saxon XSLT Engine."""
    access_id = kwargs.get("access_id")
    access_secret = kwargs.get("access_secret")
    bucket = kwargs.get("bucket")
    dest_prefix = kwargs.get("destination_prefix")
    source_prefix = kwargs.get("source_prefix")
    if kwargs.get("dag"):
        run_id = kwargs.get("dag").dag_id
    else:
        run_id = "no-dag-provided"

    saxon = prepare_saxon_engine()
    transformed = etree.Element("collection")
    transformed.attrib["dag-id"] = run_id
    transformed.attrib["dag-timestamp"] = kwargs.get("timestamp", "no-timestamp-provided")
    xsl = "https://raw.github.com/{repo}/{branch}/{filename}".format(
        repo=kwargs.get("xsl_repository", "tulibraries/aggregator_mdx"),
        branch=kwargs.get("xsl_branch", "main"),
        filename=kwargs.get("xsl_filename")
    )

    for s3_key in process.list_s3_content(bucket, access_id, access_secret, source_prefix):
        logging.info("Transforming File %s", s3_key)
        s3_content = process.get_s3_content(bucket, s3_key, access_id, access_secret)
        s3_xml = etree.fromstring(s3_content)
        for record in s3_xml.iterchildren():
            record_id = record.get("airflow-record-id")
            logging.info("Transforming Record %s", record_id)
            result_str = subprocess.check_output(["java", "-jar", saxon, "-xsl:" + xsl, "-s:-"], input=etree.tostring(record, encoding="utf-8"))
            result = etree.fromstring(result_str)
            result.attrib["airflow-record-id"] = record_id
            transformed.append(result)
        filename = s3_key.replace(source_prefix, dest_prefix)
        transformed_xml = etree.tostring(transformed, encoding="utf-8")
        process.generate_s3_object(transformed_xml, bucket, filename, access_id, access_secret)


def prepare_saxon_engine(saxon_jar="saxon.jar", saxon_path="/tmp/saxon/"):
    """Set up Saxon HE 9 Java Engine for XML & XSL tasks."""
    saxon_version = "9.9.1-5"
    saxon_download_sha1 = "c1f413a1b810dbf0d673ffd3b27c8829a82ac31c"

    if not os.path.exists(saxon_path + saxon_jar):
        Path(saxon_path).mkdir(parents=True, exist_ok=True)
        request_url = "https://repo1.maven.org/maven2/net/sf/saxon/Saxon-HE/"
        request_url += saxon_version + "/Saxon-HE-" + saxon_version + ".jar"
        resp = requests.get(request_url, allow_redirects=True, verify=False)
        if hashlib.sha1(resp.content).hexdigest() == saxon_download_sha1:
            open(saxon_path + saxon_jar, "wb").write(resp.content)
            os.chmod(saxon_path + saxon_jar, 0o744)
            return saxon_path + saxon_jar
        logging.fatal("SHA1 Digests do not match.")
        sys.exit()
    return saxon_path + saxon_jar



from tulflow import harvest
from os import listdir
import unittest
import httpretty

class TestDEVO265Bug(unittest.TestCase):
    """Test Class for S3 Post Wrapper."""


    @httpretty.activate
    def test_devo_265_lxml_bug(self, **kwargs):
        """Test Calling handling XML Element to String."""

        fixture_path = "./tests/fixtures/devo-265-failing-record.xml"
        
        body = open(fixture_path, "r").read()


        httpretty.register_uri(
            httpretty.GET,
            "http://test/combine/oai",
            body=body
        )

        kwargs["oai_endpoint"] = "http://test/combine/oai"
        kwargs["harvest_params"] = {
                "metadataPrefix": "marc21",
                "included_sets": ["blacklight"],
                "from": 2019-10-21,
                "until": None
        }

        with self.assertLogs() as log:
            response = harvest.harvest_oai(**kwargs)
            processed = harvest.process_xml(response, harvest.write_log, "test-dir", **kwargs)



"""Generic Airflow Tasks Functions, Abstracted for Reuse."""
import re
import pprint
from airflow.hooks.base_hook import BaseHook
from airflow.contrib.operators.slack_webhook_operator import SlackWebhookOperator
from airflow.operators.http_operator import SimpleHttpOperator
from airflow.operators.python_operator import PythonOperator
from tulflow.solr_api_utils import SolrApiUtils

PP = pprint.PrettyPrinter(indent=4)


def execute_slackpostonfail(context, conn_id="AIRFLOW_CONN_SLACK_WEBHOOK", message=None):
    """Task Method to Post Failed DAG or Task Completion on Slack."""
    conn = BaseHook.get_connection(conn_id)
    task_instance = context.get("task_instance")
    log_url = task_instance.log_url
    task_id = task_instance.task_id
    dag_id = task_instance.dag_id
    task_date = context.get("execution_date")
    if not message:
        message = "Task failed: {} {} {} {}".format(dag_id, task_id, task_date, log_url)

    slack_post = SlackWebhookOperator(
        task_id="slackpostonfail",
        http_conn_id=conn_id,
        webhook_token=conn.password,
        message=":poop: " + message,
        username="airflow",
        dag=context.get("dag")
        )

    return slack_post.execute(context=context)


def execute_slackpostonsuccess(context, conn_id="AIRFLOW_CONN_SLACK_WEBHOOK", message=None):
    """Task Method to Post Successful DAG or Task Completion on Slack."""
    conn = BaseHook.get_connection(conn_id)
    task_instance = context.get("task_instance")
    log_url = task_instance.log_url
    task_id = task_instance.task_id
    dag_id = task_instance.dag_id
    task_date = context.get("execution_date")
    if not message:
        message = "DAG success: {} {} {} {}".format(dag_id, task_id, task_date, log_url)

    slack_post = SlackWebhookOperator(
        task_id="slackpostonsuccess",
        http_conn_id=conn_id,
        webhook_token=conn.password,
        message=":partygritty: " + message,
        username="airflow",
        trigger_rule="all_success",
        dag=context.get("dag")
    )

    return slack_post.execute(context=context)


def create_sc_collection(dag, sc_conn_id, sc_coll_name, sc_coll_repl, sc_configset_name):
    """Creates a new SolrCloud Collection."""
    task_instance = SimpleHttpOperator(
        task_id="create_collection",
        method="GET",
        http_conn_id=sc_conn_id,
        endpoint="solr/admin/collections",
        data={
            "action": "CREATE",
            "name": sc_coll_name,
            "numShards": "1",
            "replicationFactor": sc_coll_repl,
            "maxShardsPerNode": "1",
            "collection.configName": sc_configset_name
        },
        headers={},
        dag=dag,
        log_response=True
    )

    return task_instance


def swap_sc_alias(dag, sc_conn_id, sc_coll_name, sc_configset_name):
    """Create or point an existing SolrCloud Alias to an existing SolrCloud Collection."""
    task_instance = SimpleHttpOperator(
        task_id="solr_alias_swap",
        method="GET",
        http_conn_id=sc_conn_id,
        endpoint="solr/admin/collections",
        data={
            "action": "CREATEALIAS",
            "name": sc_configset_name,
            "collections": [sc_coll_name]
            },
        headers={},
        dag=dag,
        log_response=True
    )

    return task_instance

def refresh_sc_collection_for_alias(dag, sc_conn, sc_coll_name, sc_alias, configset,numShards=None, replicationFactor=None, maxShardsPerNode=None):
    """Removes collection from alias, deletes collection, creates new collection & adds to alias"""
    task_instance = PythonOperator(
        task_id="refresh_sc_collection_for_alias",
        python_callable=SolrApiUtils.remove_and_recreate_collection_from_alias,
        op_kwargs={
            "collection": sc_coll_name,
            "configset": configset,
            "alias": sc_alias,
            "solr_url": sc_conn.host,
            "solr_port": sc_conn.port,
            "solr_auth_user": sc_conn.login or "",
            "solr_auth_pass": sc_conn.password or "",
            "numShards": numShards,
            "replicationFactor": replicationFactor,
            "maxShardsPerNode": maxShardsPerNode
        },
        dag=dag
    )
    return task_instance

def get_solr_url(conn, core):
    """  Generates a solr url from  passed in connection and core.

    Parameters:
        conn (airflow.models.connection): Connection object representing solr we index to.
        core (str)  The solr collection or configuration  to use.

    Returns:
        solr_url (str): A solr URL.

    """
    solr_url = conn.host

    if not re.match("^http", solr_url):
        solr_url = "http://" + solr_url

    if conn.port:
        solr_url += ":" + str(conn.port)

    solr_url += "/solr/" + core

    return solr_url

def conditionally_trigger(context, dag_run_obj):
    """This function decides whether or not to Trigger the remote DAG"""
    c_p = context['params']['condition_param']
    print("Controller DAG : conditionally_trigger = {}".format(c_p))
    if context['params']['condition_param']:
        dag_run_obj.payload = context['params']
        PP.pprint(dag_run_obj.payload)
        return dag_run_obj



"""Tests suite for tulflow harvest (Functions for harvesting OAI in Airflow Tasks)."""
from datetime import datetime
import hashlib
import unittest
import boto3
from lxml import etree
from moto import mock_s3
from unittest import mock
from unittest.mock import patch
from airflow.hooks.S3_hook import S3Hook
from airflow.models import DAG
from airflow.utils import timezone
from lxml import etree
import httpretty
from tulflow import harvest
from types import SimpleNamespace
from moto import mock_s3

DEFAULT_DATE = timezone.datetime(2019, 8, 16)
NS = {
    "marc21": "http://www.loc.gov/MARC21/slim",
    "oai": "http://www.openarchives.org/OAI/2.0/"
    }

lizards = """
<OAI-PMH xmlns="http://www.openarchives.org/OAI/2.0/"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xsi:schemaLocation="http://www.openarchives.org/OAI/2.0/ http://www.openarchives.org/OAI/2.0/OAI-PMH.xsd">
    <responseDate>2019-08-30T13:46:14Z</responseDate>
    <request verb="ListRecords" set="dpla_test">http://10.5.0.10/combine/oai</request>
    <ListRecords>
        <record>
            <header>
                <identifier>oai:lizards</identifier>
                <datestamp>2019-08-30T13:45:28Z</datestamp>
                <setSpec>dpla_test</setSpec>
            </header>
            <metadata>
                <oai_dc:dc xmlns:dc="http://purl.org/dc/elements/1.1/"
                    xmlns:dcterms="http://purl.org/dc/terms/"
                    xmlns:oai_dc="http://www.openarchives.org/OAI/2.0/oai_dc/">
                    <dcterms:title>lizards</dcterms:title>
                </oai_dc:dc>
            </metadata>
        </record>
    </ListRecords>
</OAI-PMH>
"""

animals = """
<OAI-PMH xmlns="http://www.openarchives.org/OAI/2.0/"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xsi:schemaLocation="http://www.openarchives.org/OAI/2.0/ http://www.openarchives.org/OAI/2.0/OAI-PMH.xsd">
    <responseDate>2019-08-30T13:46:14Z</responseDate>
    <request verb="ListRecords" set="dpla_test">http://10.5.0.10/combine/oai</request>
    <ListRecords>
        <record>
            <header>
                <identifier>oai:cats</identifier>
                <datestamp>2019-08-30T13:45:28Z</datestamp>
                <setSpec>dpla_test</setSpec>
            </header>
            <metadata>
                <oai_dc:dc xmlns:dc="http://purl.org/dc/elements/1.1/"
                    xmlns:dcterms="http://purl.org/dc/terms/"
                    xmlns:oai_dc="http://www.openarchives.org/OAI/2.0/oai_dc/">
                    <dcterms:title>cats</dcterms:title>
                </oai_dc:dc>
            </metadata>
        </record>
        <record>
            <header>
                <identifier>oai:kittens</identifier>
                <datestamp>2019-08-30T13:45:28Z</datestamp>
                <setSpec>dpla_test</setSpec>
            </header>
            <metadata>
                <oai_dc:dc xmlns:dc="http://purl.org/dc/elements/1.1/"
                    xmlns:dcterms="http://purl.org/dc/terms/"
                    xmlns:oai_dc="http://www.openarchives.org/OAI/2.0/oai_dc/">
                    <dcterms:title>kittens</dcterms:title>
                </oai_dc:dc>
            </metadata>
        </record>
    </ListRecords>
</OAI-PMH>
"""

marc = """
<OAI-PMH xmlns="http://www.openarchives.org/OAI/2.0/"
         xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
         xsi:schemaLocation="http://www.openarchives.org/OAI/2.0/ http://www.openarchives.org/OAI/2.0/OAI-PMH.xsd">
    <responseDate>2019-11-08T13:43:00Z</responseDate>
    <request verb="ListRecords" metadataPrefix="marc21" set="blacklight">https://na02.alma.exlibrisgroup.com/view/oai/01TULI_INST/request</request>
    <ListRecords>
        <record>
            <header>
                <identifier>oai:alma.01TULI_INST:991000000269703811</identifier>
                <datestamp>2019-07-15T15:17:33Z</datestamp>
                <setSpec>blacklight</setSpec>
                <setSpec>blacklight_qa</setSpec>
                <setSpec>rapid_print_books</setSpec>
            </header>
            <metadata>
                <record xmlns="http://www.loc.gov/MARC21/slim" xsi:schemaLocation="http://www.loc.gov/MARC21/slim http://www.loc.gov/standards/marcxml/schema/MARC21slim.xsd">
                    <leader>01407nam a2200445 4500</leader>
                    <controlfield tag="005">20190715090942.0</controlfield>
                    <controlfield tag="008">690326s1969 nju b 000 0 eng </controlfield>
                    <controlfield tag="001">991000000269703811</controlfield>
                    <datafield tag="010" ind1=" " ind2=" ">
                        <subfield code="a">68020157</subfield>
                    </datafield>
                    <datafield tag="035" ind1=" " ind2=" ">
                        <subfield code="a">(PPT)b10000276-01tuli_inst</subfield>
                    </datafield>
                    <datafield tag="040" ind1=" " ind2=" ">
                        <subfield code="a">DLC</subfield>
                        <subfield code="b">eng</subfield>
                        <subfield code="c">DLC</subfield>
                        <subfield code="d">PPT</subfield>
                    </datafield>
                    <datafield tag="090" ind1=" " ind2=" ">
                        <subfield code="a">HM131.E85</subfield>
                    </datafield>
                    <datafield tag="100" ind1="1" ind2=" ">
                        <subfield code="a">Etzioni, Amitai.</subfield>
                        <subfield code="0">http://id.loc.gov/authorities/names/n79089329</subfield>
                    </datafield>
                    <datafield tag="245" ind1="1" ind2="0">
                        <subfield code="a">Readings on modern organizations.</subfield>
                    </datafield>
                </record>
            </metadata>
        </record>
        <record>
            <header status="deleted">
                <identifier>oai:alma.01TULI_INST:991000000939703811</identifier>
                <datestamp>2018-04-02T21:02:12Z</datestamp>
                <setSpec>blacklight</setSpec>
            </header>
        </record>
        <record>
            <header status="deleted">
                <identifier>oai:alma.01TULI_INST:991000000939703812</identifier>
                <datestamp>2018-04-02T21:02:12Z</datestamp>
                <setSpec>blacklight</setSpec>
            </header>
        </record>
    </ListRecords>
</OAI-PMH>
"""

marc_single = """
<record xmlns="http://www.openarchives.org/OAI/2.0/"
 xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
 xsi:schemaLocation="http://www.openarchives.org/OAI/2.0/ http://www.openarchives.org/OAI/2.0/OAI-PMH.xsd">
    <header>
        <identifier>oai:alma.01TULI_INST:991000000269703811</identifier>
        <datestamp>2019-07-15T15:17:33Z</datestamp>
        <setSpec>blacklight</setSpec>
        <setSpec>blacklight_qa</setSpec>
        <setSpec>rapid_print_books</setSpec>
    </header>
    <metadata>
        <record xmlns="http://www.loc.gov/MARC21/slim" xsi:schemaLocation="http://www.loc.gov/MARC21/slim http://www.loc.gov/standards/marcxml/schema/MARC21slim.xsd">
            <leader>01407nam a2200445 4500</leader>
            <controlfield tag="005">20190715090942.0</controlfield>
            <controlfield tag="008">690326s1969 nju b 000 0 eng </controlfield>
            <controlfield tag="001">991000000269703811</controlfield>
            <datafield tag="010" ind1=" " ind2=" ">
                <subfield code="a">68020157</subfield>
            </datafield>
            <datafield tag="035" ind1=" " ind2=" ">
                <subfield code="a">(PPT)b10000276-01tuli_inst</subfield>
            </datafield>
            <datafield tag="040" ind1=" " ind2=" ">
                <subfield code="a">DLC</subfield>
                <subfield code="b">eng</subfield>
                <subfield code="c">DLC</subfield>
                <subfield code="d">PPT</subfield>
            </datafield>
            <datafield tag="090" ind1=" " ind2=" ">
                <subfield code="a">HM131.E85</subfield>
            </datafield>
            <datafield tag="100" ind1="1" ind2=" ">
                <subfield code="a">Etzioni, Amitai.</subfield>
                <subfield code="0">http://id.loc.gov/authorities/names/n79089329</subfield>
            </datafield>
            <datafield tag="245" ind1="1" ind2="0">
                <subfield code="a">Readings on modern organizations.</subfield>
            </datafield>
        </record>
    </metadata>
</record>
"""

lookup = """child_id,parent_id,parent_xml
991000000269703811,9910367273103811,"<datafield>test</datafield>||<datafield>9910367273103811</datafield>"
"""

listSets = """
<?xml version="1.0" encoding="UTF-8"?>
<OAI-PMH
    xmlns="http://www.openarchives.org/OAI/2.0/"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xsi:schemaLocation="http://www.openarchives.org/OAI/2.0/ http://www.openarchives.org/OAI/2.0/OAI-PMH.xsd">
    <responseDate>2019-11-11T21:17:32Z</responseDate>
    <request verb="ListSets">http://digitalcollections.plymouth.edu/oai/oai.php</request>
    <ListSets>
        <set>
            <setSpec>a</setSpec>
            <setName>A</setName>
            <setDescription>
                <oai_dc:dc
                    xmlns:oai_dc="http://www.openarchives.org/OAI/2.0/oai_dc/"
                    xmlns:dc="http://purl.org/dc/elements/1.1/"
                    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                    xsi:schemaLocation="http://www.openarchives.org/OAI/2.0/oai_dc/ http://www.openarchives.org/OAI/2.0/oai_dc.xsd" />
            </setDescription>
        </set>
        <set>
            <setSpec>b</setSpec>
            <setName>B</setName>
            <setDescription>
                <oai_dc:dc
                    xmlns:oai_dc="http://www.openarchives.org/OAI/2.0/oai_dc/"
                    xmlns:dc="http://purl.org/dc/elements/1.1/"
                    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                    xsi:schemaLocation="http://www.openarchives.org/OAI/2.0/oai_dc/ http://www.openarchives.org/OAI/2.0/oai_dc.xsd" />
            </setDescription>
        </set>
        <set>
            <setSpec>x</setSpec>
            <setName>X</setName>
            <setDescription>
                <oai_dc:dc
                    xmlns:oai_dc="http://www.openarchives.org/OAI/2.0/oai_dc/"
                    xmlns:dc="http://purl.org/dc/elements/1.1/"
                    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                    xsi:schemaLocation="http://www.openarchives.org/OAI/2.0/oai_dc/ http://www.openarchives.org/OAI/2.0/oai_dc.xsd" />
            </setDescription>
        </set>
        <set>
            <setSpec>y</setSpec>
            <setName>Y</setName>
            <setDescription>
                <oai_dc:dc
                    xmlns:oai_dc="http://www.openarchives.org/OAI/2.0/oai_dc/"
                    xmlns:dc="http://purl.org/dc/elements/1.1/"
                    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                    xsi:schemaLocation="http://www.openarchives.org/OAI/2.0/oai_dc/ http://www.openarchives.org/OAI/2.0/oai_dc.xsd" />
            </setDescription>
        </set>
        <set>
            <setSpec>z</setSpec>
            <setName>Z</setName>
            <setDescription>
                <oai_dc:dc
                    xmlns:oai_dc="http://www.openarchives.org/OAI/2.0/oai_dc/"
                    xmlns:dc="http://purl.org/dc/elements/1.1/"
                    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                    xsi:schemaLocation="http://www.openarchives.org/OAI/2.0/oai_dc/ http://www.openarchives.org/OAI/2.0/oai_dc.xsd" />
            </setDescription>
        </set>
    </ListSets>
</OAI-PMH>
"""


class TestDagS3Interaction(unittest.TestCase):
    """Test Class for S3 Post Wrapper."""
    @classmethod
    def setUpClass(self):
        self.dag_id = "s3_stuff"
        self.maxDiff = None


    def test_dag_s3_prefix(self):
        """Test Creating S3 Bucket ("prefix")."""
        timestamp = datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")
        prefix = harvest.dag_s3_prefix(self.dag_id, timestamp)
        self.assertEqual(prefix, "{}/{}".format(self.dag_id, timestamp))


    @mock.patch("tulflow.process.generate_s3_object")
    def test_write_push_string_to_s3(self, mock):
        """Test Writing String to S3 using our Function."""
        string = "<fooooooooo>"
        prefix = "this/thing/here"
        kwargs = {}
        kwargs["access_id"] = "puppies"
        kwargs["access_secret"] = "kittens"
        kwargs["bucket_name"] = "my-bucket"
        our_hash = hashlib.md5(string.encode("utf-8")).hexdigest()
        key = "{}/{}".format(prefix, our_hash)


        harvest.dag_write_string_to_s3(string=string, prefix=prefix, **kwargs)
        mock.assert_called_once_with(string, "my-bucket", key, "puppies", "kittens")


class TestOAIHarvestInteraction(unittest.TestCase):
    """Test Class for OAI Harvest Wrapper."""
    @classmethod
    def setUpClass(self):
        self.maxDiff = None


    def test_oai_sets_all(self, **kwargs):
        """Test that all sets are lined up for harvest."""
        kwargs["all_sets"] = True
        kwargs["included_sets"] = ["a", "b", "c"]
        kwargs["excluded_sets"] = ["x", "y", "z"]

        returned_sets = harvest.generate_oai_sets(**kwargs)
        self.assertEqual(returned_sets, [])


    def test_oai_sets_incl(self, **kwargs):
        """Test that all sets are lined up for harvest."""
        kwargs["all_sets"] = False
        kwargs["included_sets"] = ["a", "b", "c"]
        kwargs["excluded_sets"] = ["x", "y", "z"]

        returned_sets = harvest.generate_oai_sets(**kwargs)
        self.assertEqual(returned_sets, ["a", "b", "c"])


    @httpretty.activate
    def test_oai_sets_excl(self, **kwargs):
        """Test that all sets are lined up for harvest."""
        httpretty.register_uri(
            httpretty.GET,
            "http://127.0.0.1/combine/oai",
            body=listSets
        )
        kwargs["oai_endpoint"] = "http://127.0.0.1/combine/oai"
        kwargs["all_sets"] = False
        kwargs["included_sets"] = []
        kwargs["excluded_sets"] = ["x", "y", "z"]

        returned_sets = harvest.generate_oai_sets(**kwargs)
        returned_sets.sort()
        self.assertListEqual(returned_sets, ["a", "b"])


    @httpretty.activate
    def test_harvest_oai(self, **kwargs):
        """Test Calling OAI-PMH HTTP Endpoint & Returning XML String."""
        httpretty.register_uri(
            httpretty.GET,
            "http://127.0.0.1/combine/oai",
            body=lizards
        )

        kwargs["oai_endpoint"] = "http://127.0.0.1/combine/oai"
        kwargs["harvest_params"] = {
            "metadataPrefix": "generic",
            "included_sets": "dpla_test",
            "from": None,
            "until": None
        }

        records = harvest.harvest_oai(**kwargs)
        xml_output = etree.tostring(records.next().xml, pretty_print=True, encoding="utf-8")
        self.assertEqual(type(records), harvest.HarvestIterator)
        self.assertIn(b"<identifier>oai:lizards</identifier>", xml_output)
        self.assertIn(b"<setSpec>dpla_test</setSpec>", xml_output)
        self.assertIn(b"<dcterms:title>lizards</dcterms:title>", xml_output)

    @httpretty.activate
    def test_harvest_oai_no_records(self, **kwargs):
        """Test Calling OAI-PMH HTTP Endpoint & Returning XML String."""
        httpretty.register_uri(
            httpretty.GET,
            "http://127.0.0.1/combine/oai",
            body="""
            <?xml version="1.0" encoding="UTF-8"?><OAI-PMH xmlns="http://www.openarchives.org/OAI/2.0/" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.openarchives.org/OAI/2.0/ http://www.openarchives.org/OAI/2.0/OAI-PMH.xsd"><responseDate>2021-01-04T19:12:46Z</responseDate><request verb="ListRecords" metadataPrefix="oai_qdc" set="p15860coll2">http://cdm15860.contentdm.oclc.org/oai/oai.php</request><error code="noRecordsMatch">The combination of the values of the from, until, set and metadataPrefix arguments results in an empty list. </error></OAI-PMH>
            """
        )

        kwargs["oai_endpoint"] = "http://127.0.0.1/combine/oai"
        kwargs["harvest_params"] = {
            "metadataPrefix": "generic",
            "included_sets": "dpla_test",
            "from": None,
            "until": None
        }

        records = harvest.harvest_oai(**kwargs)
        self.assertEqual(records, [])

    @httpretty.activate
    def test_harvest_oai_no_metadata(self, **kwargs):
        """Test Calling OAI-PMH HTTP Endpoint & Returning XML String."""
        httpretty.register_uri(
            httpretty.GET,
            "http://127.0.0.1/combine/oai",
            body = """
<OAI-PMH xmlns="http://www.openarchives.org/OAI/2.0/"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xsi:schemaLocation="http://www.openarchives.org/OAI/2.0/ http://www.openarchives.org/OAI/2.0/OAI-PMH.xsd">
    <responseDate>2019-08-30T13:46:14Z</responseDate>
    <request verb="ListRecords" set="dpla_test">http://10.5.0.10/combine/oai</request>
    <ListRecords>
        <record>
            <header>
                <identifier>oai:lizards</identifier>
                <datestamp>2019-08-30T13:45:28Z</datestamp>
                <setSpec>dpla_test</setSpec>
            </header>
        </record>
    </ListRecords>
</OAI-PMH>
"""
        )

        kwargs["oai_endpoint"] = "http://127.0.0.1/combine/oai"
        kwargs["harvest_params"] = {
            "metadataPrefix": "generic",
            "included_sets": "dpla_test",
            "from": None,
            "until": None
        }

        records = harvest.harvest_oai(**kwargs)
        # Skips records with no metadata.
        self.assertEqual(len(list(records)), 0)

    @httpretty.activate
    def test_process_xml_dpla(self, **kwargs):
        """Test Calling handling XML Element to String."""
        httpretty.register_uri(
            httpretty.GET,
            "http://test/combine/oai",
            body=animals
        )

        kwargs["oai_endpoint"] = "http://test/combine/oai"
        kwargs["harvest_params"] = {
            "metadataPrefix": "generic",
            "included_sets": ["dpla_test"],
            "from": None,
            "until": None
        }

        with self.assertLogs() as log:
            response = harvest.harvest_oai(**kwargs)
            processed = harvest.process_xml(response, harvest.write_log, "test-dir", **kwargs)
            self.assertIn("INFO:root:OAI Records Harvested & Processed: 2", log.output)
            self.assertIn("INFO:root:OAI Records Harvest & Marked for Deletion: 0", log.output)
            self.assertEqual(processed, {"updated": 2, "deleted": 0 })


    @httpretty.activate
    def test_process_xml_alma(self, **kwargs):
        """Test Calling handling XML Element to String with Deletes."""
        httpretty.register_uri(
            httpretty.GET,
            "http://127.0.0.1/alma/oai",
            body=marc
        )

        kwargs["oai_endpoint"] = "http://127.0.0.1/alma/oai"
        kwargs["harvest_params"] = {
            "metadataPrefix": "marc21",
            "included_sets": ["blacklight"],
            "from": 2019-10-21,
            "until": None
        }

        with self.assertLogs() as log:
            response = harvest.harvest_oai(**kwargs)
            harvest.process_xml(response, harvest.write_log, "test-dir", **kwargs)
        self.assertIn("INFO:root:OAI Records Harvested & Processed: 1", log.output)
        self.assertIn("INFO:root:OAI Records Harvest & Marked for Deletion: 2", log.output)
        self.assertIn('INFO:root:<oai:collection xmlns:oai="http://www.openarchives.org/OAI/2.0/" dag-id="no-dag-provided" dag-timestamp="no-timestamp-provided"><oai:record xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" airflow-record-id="oai:alma.01TULI_INST:991000000269703811"><oai:header><oai:identifier>oai:alma.01TULI_INST:991000000269703811</oai:identifier><oai:datestamp>2019-07-15T15:17:33Z</oai:datestamp><oai:setSpec>blacklight</oai:setSpec><oai:setSpec>blacklight_qa</oai:setSpec><oai:setSpec>rapid_print_books</oai:setSpec></oai:header><oai:metadata><record xmlns="http://www.loc.gov/MARC21/slim" xsi:schemaLocation="http://www.loc.gov/MARC21/slim http://www.loc.gov/standards/marcxml/schema/MARC21slim.xsd"><leader>01407nam a2200445 4500</leader><controlfield tag="005">20190715090942.0</controlfield><controlfield tag="008">690326s1969 nju b 000 0 eng </controlfield><controlfield tag="001">991000000269703811</controlfield><datafield tag="010" ind1=" " ind2=" "><subfield code="a">68020157</subfield></datafield><datafield tag="035" ind1=" " ind2=" "><subfield code="a">(PPT)b10000276-01tuli_inst</subfield></datafield><datafield tag="040" ind1=" " ind2=" "><subfield code="a">DLC</subfield><subfield code="b">eng</subfield><subfield code="c">DLC</subfield><subfield code="d">PPT</subfield></datafield><datafield tag="090" ind1=" " ind2=" "><subfield code="a">HM131.E85</subfield></datafield><datafield tag="100" ind1="1" ind2=" "><subfield code="a">Etzioni, Amitai.</subfield><subfield code="0">http://id.loc.gov/authorities/names/n79089329</subfield></datafield><datafield tag="245" ind1="1" ind2="0"><subfield code="a">Readings on modern organizations.</subfield></datafield></record></oai:metadata></oai:record></oai:collection>', log.output)
        # assert multiple deletions get added as expected
        self.assertIn('INFO:root:<oai:collection xmlns:oai="http://www.openarchives.org/OAI/2.0/" dag-id="no-dag-provided" dag-timestamp="no-timestamp-provided"><oai:record airflow-record-id="oai:alma.01TULI_INST:991000000939703811"><oai:header status="deleted"><oai:identifier>oai:alma.01TULI_INST:991000000939703811</oai:identifier><oai:datestamp>2018-04-02T21:02:12Z</oai:datestamp><oai:setSpec>blacklight</oai:setSpec></oai:header></oai:record><oai:record airflow-record-id="oai:alma.01TULI_INST:991000000939703812"><oai:header status="deleted"><oai:identifier>oai:alma.01TULI_INST:991000000939703812</oai:identifier><oai:datestamp>2018-04-02T21:02:12Z</oai:datestamp><oai:setSpec>blacklight</oai:setSpec></oai:header></oai:record></oai:collection>', log.output)


    @mock_s3
    def test_perform_xml_lookup_with_cache(self, **kwargs):
        """Test Calling handling XML Element to String with Deletes."""
        kwargs["access_id"] = "cats"
        kwargs["access_secret"] = "dogs"
        kwargs["bucket_name"] = "alma-test"
        kwargs["lookup_key"] = "sc_catalog_pipeline/0000-00-00/lookup.tsv"

        conn = boto3.client(
            "s3",
            aws_access_key_id=kwargs.get("access_id"),
            aws_secret_access_key=kwargs.get("access_secret")
        )
        conn.create_bucket(Bucket=kwargs.get("bucket_name"))
        conn.put_object(Bucket=kwargs.get("bucket_name"), Key=kwargs.get("lookup_key"), Body=lookup)
        marc_xml = etree.fromstring(marc_single)
        with self.assertLogs() as log:
            parser = harvest.perform_xml_lookup_with_cache()
            resp_xml = parser(marc_xml, **kwargs)
        self.assertIn("INFO:root:Reading in Record 991000000269703811", log.output)
        self.assertIn("INFO:root:Child XML record found 991000000269703811", log.output)
        self.assertIn(b"<datafield>test</datafield>", etree.tostring(resp_xml))
        self.assertIn(b"<datafield>9910367273103811</datafield>", etree.tostring(resp_xml))


    @mock.patch("tulflow.harvest.harvest_oai")
    @mock.patch("tulflow.harvest.dag_s3_prefix")
    @mock.patch("tulflow.harvest.process_xml")
    def test_oai_to_s3_harvest(self, mock_process, mock_prefix, mock_harvest, **kwargs):
        """Test oai_to_s3 wraps harvest_oai function."""
        dag = DAG(dag_id="test_slacksuccess", start_date=DEFAULT_DATE)
        kwargs["oai_endpoint"] = "http://test/combine/oai"
        kwargs["metadataPrefix"] = "blergh"
        kwargs["included_sets"] = ["set"]
        kwargs["from"] = "from"
        kwargs["until"] = "until"
        kwargs["dag"] = dag
        kwargs["timestamp"] = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

        harvest.oai_to_s3(**kwargs)
        self.assertTrue(mock_harvest.called)
        self.assertTrue(mock_prefix.called)
        self.assertTrue(mock_process.called)


    @mock.patch("tulflow.harvest.harvest_oai")
    @mock.patch("tulflow.harvest.dag_s3_prefix")
    @mock.patch("tulflow.harvest.process_xml")
    def test_oai_to_s3_harvest_return_value_mult_set(self, mock_process, mock_prefix, mock_harvest, **kwargs):
        """Test oai_to_s3 wraps harvest_oai function."""
        dag = DAG(dag_id="test_slacksuccess", start_date=DEFAULT_DATE)
        kwargs["oai_endpoint"] = "http://test/combine/oai"
        kwargs["metadataPrefix"] = "blergh"
        kwargs["included_sets"] = ["set1", "set2"]
        kwargs["from"] = "from"
        kwargs["until"] = "until"
        kwargs["dag"] = dag
        kwargs["timestamp"] = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        mock_process.return_value = {"updated": 2, "deleted": 0 }
        actual = harvest.oai_to_s3(**kwargs)
        self.assertEqual(actual, {"updated": 4, "deleted": 0, "sets_with_no_records": []})

    @mock.patch("tulflow.harvest.harvest_oai")
    @mock.patch("tulflow.harvest.dag_s3_prefix")
    @mock.patch("tulflow.harvest.process_xml")
    def test_oai_to_s3_harvest_not_process_empty_data(self, mock_process, mock_prefix, mock_harvest, **kwargs):
        """Test oai_to_s3 does not process empty data"""
        dag = DAG(dag_id="test_slacksuccess", start_date=DEFAULT_DATE)
        kwargs["oai_endpoint"] = "http://test/combine/oai"
        kwargs["metadataPrefix"] = "blergh"
        kwargs["included_sets"] = ["set1", "set2"]
        kwargs["from"] = "from"
        kwargs["until"] = "until"
        kwargs["dag"] = dag
        kwargs["timestamp"] = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

        mock_harvest.return_value = []
        mock_process.return_value = {"updated": 0, "deleted": 0 }
        actual = harvest.oai_to_s3(**kwargs)
        self.assertFalse(mock_process.called)
        self.assertEqual(actual, {"updated": 0, "deleted": 0, "sets_with_no_records": ["set1", "set2"]})

    @mock.patch("tulflow.harvest.harvest_oai")
    @mock.patch("tulflow.harvest.dag_s3_prefix")
    @mock.patch("tulflow.harvest.process_xml")
    def test_oai_to_s3_harvest_return_value_no_set(self, mock_process, mock_prefix, mock_harvest, **kwargs):
        """Test oai_to_s3 wraps harvest_oai function."""
        dag = DAG(dag_id="test_slacksuccess", start_date=DEFAULT_DATE)
        kwargs["oai_endpoint"] = "http://test/combine/oai"
        kwargs["metadataPrefix"] = "blergh"
        kwargs["from"] = "from"
        kwargs["until"] = "until"
        kwargs["all_sets"] = True
        kwargs["dag"] = dag
        kwargs["timestamp"] = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        mock_process.return_value = {"updated": 2, "deleted": 0 }
        actual = harvest.oai_to_s3(**kwargs)
        self.assertEqual(actual, {"updated": 2, "deleted": 0, "sets_with_no_records": [] })




"""Submodule for interating with SolrCloud API *not* through Airflow Connections."""
import logging
from collections import defaultdict
from re import split as resplit
import requests


class SolrApiUtils():
    """Class for interacting with SolrCloud API over HTTP."""

    @classmethod
    def remove_and_recreate_collection_from_alias(cls, collection, alias, solr_url, configset="_default", solr_port=None, solr_auth_user=None, solr_auth_pass=None, numShards=None, replicationFactor=None, maxShardsPerNode=None):
        """Method to remove & re-add collection to SolrCloud Alias."""
        url = solr_url + (f":{solr_port}" if solr_port else "")
        logging.info("Trying %s", url)
        sc = cls(url, auth_user=solr_auth_user, auth_pass=solr_auth_pass)
        sc.remove_collection_from_alias(collection=collection, alias=alias)
        sc.delete_collection(collection)
        create_params = {"collection": collection, "configset": configset, "numShards": numShards, "replicationFactor": replicationFactor, "maxShardsPerNode": maxShardsPerNode}
        cleaned_create_params = {k: v for k, v in create_params.items() if v is not None}
        sc.create_collection(**cleaned_create_params)
        sc.add_collection_to_alias(collection=collection, alias=alias)

    def __init__(self, solr_url, auth_user=None, auth_pass=None):
        self.solr_url = solr_url
        self.auth_user = auth_user
        self.auth_pass = auth_pass

    def get_from_solr_api(self, path):
        """Issue HTTP Get Request to SolrCloud API."""
        url = self.solr_url + path
        logging.info("Requesting %s", url)
        if self.auth_user and self.auth_pass:
            return requests.get(url, auth=(self.auth_user, self.auth_pass))
        return requests.get(url)

    def get_configsets(self):
        """Issue HTTP Get Request to SolrCloud API to retrieve ConfigSets List."""
        if not hasattr(self, "configsets"):
            json_response = self.get_from_solr_api("/api/cluster/configs?omitHeader=true").json()
            try:
                self.configsets = json_response['configSets']
            except KeyError as solr_error:
                msg = json_response.get('error', {}).get("msg", "Unknown error from Solr")
                raise KeyError(f"{solr_error}: caused by\n{msg}")
        return self.configsets

    def most_recent_configsets(self):
        """Issue HTTP Get Request to SolrCloud API to retrieve Recent ConfigSets."""
        configs = defaultdict(set)
        for configset in self.get_configsets():
            split = resplit(r'-(\d+)', configset)
            if len(split) > 1:
                config, version, *_ = split
                configs[config].add(int(version))
        return [f"{configset}-{max(versions)}" for (configset, versions) in configs.items()]

    def get_collections(self):
        """Issue HTTP Get Request to SolrCloud API to retrieve Collections List."""
        if not hasattr(self, "collections"):
            self.collections = self.get_from_solr_api("/solr/admin/collections?action=List").json()['collections']
        return self.collections

    def delete_collection(self, collection):
        """Issue HTTP Get Request to SolrCloud API to delete a Collection."""
        if self.collection_exists(collection):
            path = "/solr/admin/collections?action=DELETE&name=" + collection
            response = self.get_from_solr_api(path)
            try:
                response.raise_for_status()
            except requests.exceptions.HTTPError as http_error:
                msg = response.json().get('error', {}).get("msg", "Unknown error from Solr")
                raise requests.exceptions.HTTPError(f"{http_error} caused by\n{msg}")
            self.__unsetattr("collections")
            logging.info("Collection %s deleted", collection)
        else:
            logging.info("Collection %s did not exist. Continuing without error", collection)

    def create_collection(self, collection, configset="_default", numShards=1, replicationFactor=1, maxShardsPerNode=1):
        """Create a SolrCloud collection (not through Airflow Task; that is in tasks.py)"""
        path = f"/solr/admin/collections?action=CREATE&name={collection}&collection.configName={configset}&numShards={numShards}&replicationFactor={replicationFactor}&maxShardsPerNode={maxShardsPerNode}"
        response = self.get_from_solr_api(path)
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as http_error:
            msg = response.json().get('error', {}).get("msg", "Unknown error from Solr")
            raise requests.exceptions.HTTPError(f"{http_error} caused by\n{msg}")
        self.__unsetattr("collections")
        logging.info("Collection %s created", collection)

    def collection_exists(self, collection):
        """Check if a SolrCloud Collection already exists."""
        return collection in self.get_collections()

    def get_aliases(self):
        """Issue HTTP Get Request to SolrCloud API to retrieve Aliases List."""
        if not hasattr(self, "aliases"):
            self.aliases = self.get_from_solr_api("/solr/admin/collections?action=ListAliases").json()['aliases']
        return self.aliases

    def create_or_modify_alias_and_set_collections(self, alias, collections):
        """Issue HTTP Get Request to SolrCloud API to update an Alias' Collections"""
        collectionlist = ",".join(collections)
        path = f"/solr/admin/collections?action=CREATEALIAS&name={alias}&collections={collectionlist}"
        response = self.get_from_solr_api(path)
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as http_error:
            msg = response.json().get('error', {}).get("msg", "Unknown error from Solr")
            raise requests.exceptions.HTTPError(f"{http_error} caused by\n{msg}")
        logging.info("%s is now an alias for collections %s", alias, collectionlist)
        # Force HTTP call for updated state of aliases
        self.__unsetattr("aliases")

    def get_alias_collections(self, alias):
        """Issue HTTP Get Request to SolrCloud API to get Collections behind an Alias."""
        if self.alias_exists(alias):
            return self.get_aliases().get(alias).split(",")
        raise LookupError(f"{alias} is not an existing alias")

    def get_alias_collections_without_init(self, alias):
        """Get list of aliases without the collection used to initialize an alias"""
        collections = self.get_alias_collections(alias)
        return self.filter_init_collection(collections)


    def is_collection_in_alias(self, collection, alias):
        """Issue HTTP Get Request to SolrCloud API to see if Collection is behind an Alias."""
        return collection in self.get_alias_collections(alias)

    def alias_exists(self, alias):
        """Check if a SolrCloud Alias Exists."""
        return alias in self.get_aliases()

    def remove_collection_from_alias(self, collection, alias):
        """Remove a SolrCloud Collection from a given Alias."""
        if self.is_collection_in_alias(collection, alias):
            collections = self.get_alias_collections(alias)
            collections.remove(collection)
            if len(collections) > 0:
                self.create_or_modify_alias_and_set_collections(alias=alias, collections=collections)
                self.__unsetattr("aliases")
            else:
                raise ValueError("Cannot delete only collection from alias")
        else:
            logging.info("Collection %s not already in alias %s. Moving on w/o error", collection, alias)

    def add_collection_to_alias(self, collection, alias):
        """Add a SolrCloud Collection to an Existing Alias."""
        collections = self.get_alias_collections_without_init(alias)
        collections.append(collection)
        self.create_or_modify_alias_and_set_collections(alias=alias, collections=collections)
        self.__unsetattr("aliases")

    def filter_init_collection(self, collections_list, init_collection_name=None):
        """Remove initial dummy collection added to create alias"""
        if not init_collection_name:
            test = lambda collection: not collection.endswith("-init")
        else:
            test = lambda collection: (collection != init_collection_name)
        return [collection for collection in collections_list if test(collection)]



    def __unsetattr(self, attr):
        if hasattr(self, attr):
            delattr(self, attr)



"""Generic Data (primarily XML) Processing Functions, Abstracted for Reuse."""
import io
import logging
import sys
import tarfile
from lxml import etree
from botocore.exceptions import ClientError
import boto3
import requests

NS = {
    "marc21": "http://www.loc.gov/MARC21/slim",
    "oai": "http://www.openarchives.org/OAI/2.0/"
    }

etree.register_namespace("marc21", "http://www.loc.gov/MARC21/slim")
etree.register_namespace("oai", "http://www.openarchives.org/OAI/2.0")

LOGGER = logging.getLogger('tulflow_process')
PARSER = etree.XMLParser(remove_blank_text=True)

def s3_client(access_id, access_secret):
    kwargs = {}
    if access_id:
        kwargs['aws_access_key_id'] = access_id

    if access_secret:
        kwargs['aws_secret_access_key'] = access_secret

    return boto3.client(
        "s3",
        "us-east-1",
        **kwargs
    )


def add_marc21xml_root_ns(data_in):
    """Given an ALMASFTP XML Collection document as bytes,
    Convert it to lxml.etree.Element & inject MARC21 as default namespace."""
    source_xml = etree.fromstring(data_in, parser=PARSER)
    if (not source_xml.attrib.get("xmlns")) and ("{http://www.loc.gov/MARC21/slim}" not in source_xml.tag):
        source_xml.attrib["xmlns"] = "http://www.loc.gov/MARC21/slim"
    # The reason this is here is to catch encoding errors early.
    source_xml = etree.fromstring(etree.tostring(source_xml, encoding="utf-8"), parser=PARSER)
    return source_xml


def expand_alma_sftp_tarball(key, source_obj):
    """Given an AlmaSFTP S3 bytestream, expand and return XML file."""
    source_tar = tarfile.open(fileobj=io.BytesIO(source_obj), mode="r:gz")
    if len(source_tar.getmembers()) == 0:
        LOGGER.error("S3 Object is empty.")
        LOGGER.error(key)
        return None

    if len(source_tar.getmembers()) > 1:
        LOGGER.error("S3 Object has more than 1 member, which is unexpected.")
        LOGGER.error(key)
        return None

    for item in source_tar:
        return source_tar.extractfile(item).read()


def get_record_001(record):
    """Given a MARC/XML record (lxml.etree.Element), validate & return OO1 text."""
    record_ids = record.xpath("marc21:controlfield[@tag='001']", namespaces=NS)

    if record_ids == [] or record_ids[0].text is None:
        LOGGER.error("Record without an 001 MMS Identifier:")
        LOGGER.error(str(etree.tostring(record, encoding="utf8")))
        return None

    if len(record_ids) > 1:
        LOGGER.error("Record with multiple 001 MMS Identifiers:")
        LOGGER.error(str(etree.tostring(record, encoding="utf8")))
        return None

    return record_ids[0].text


def generate_bw_parent_field(parent_id):
    """Generates our Parent ID MARC/XML field inserted into the relevant Child Records."""
    new_field = etree.Element("{http://www.loc.gov/MARC21/slim}datafield")
    new_field.set("ind1", " ")
    new_field.set("ind2", " ")
    new_field.set("tag", "ADF")
    subfield = etree.SubElement(new_field, "{http://www.loc.gov/MARC21/slim}subfield")
    subfield.set("code", "a")
    subfield.text = parent_id
    return new_field

def get_github_content(repository, filename, branch="main"):
    """Get the contents of GitHub file."""
    raw_url = "https://raw.github.com/{repo}/{branch}/{filename}".format(
        repo=repository,
        branch=branch,
        filename=filename
    )
    try:
        resp = requests.get(raw_url)
        resp.raise_for_status()
        return resp.content
    except requests.exceptions.RequestException as error:
        logging.error(error)
        sys.exit(1)

def remove_s3_object(bucket, key, access_id, access_secret):
    """Removes an S3 object."""
    try:
        s3_client(access_id, access_secret).delete_object(Bucket=bucket, Key=key)
    except ClientError as error:
        LOGGER.error(error)


def get_s3_content(bucket, key, access_id, access_secret):
    """Get the contents of S3 object located at given S3 Key."""
    try:
        response = s3_client(access_id, access_secret).get_object(Bucket=bucket, Key=key)
        body = response['Body'].read()
        return body
    except ClientError as error:
        LOGGER.error(error)
        return None


def list_s3_content(bucket, access_id, access_secret, prefix=""):
    """Get a list of S3 objects located in a Bucket at the given Prefix"""
    try:
        response = s3_client(access_id, access_secret).list_objects(Bucket=bucket, Prefix=prefix)
        objects = []
        if response.get("Contents"):
            for key in response['Contents']:
                objects.append(key['Key'])
        return objects
    except ClientError as error:
        LOGGER.error(error)
        return None


def generate_s3_object(body, bucket, key, access_id, access_secret):
    try:
        s3_client(access_id, access_secret).put_object(Bucket=bucket, Key=key, Body=body)
    except ClientError as error:
        LOGGER.error(error)



"""Tests suite for tulflow.transform (functions for transforming XML or JSON in Airflow Tasks)."""
import unittest
import boto3
from lxml import etree
from moto import mock_s3
from tulflow import transform
import logging
from unittest.mock import patch

class TestXSLTransform(unittest.TestCase):
    """Test Class for functions that transform XML from S3 with XSL."""
    maxDiff = None
    kwargs = {
        "source_prefix": "dpla_test/new-updated-filtered",
        "destination_prefix": "dpla_test/transformed",
        "bucket": "tulib-airflow-test",
        "schematron_filename": "transforms/dplah.xsl",
        "access_id": "kittens",
        "access_secret": "puppies"
    }

    def setUp(self):
        transform.prepare_saxon_engine()

    @mock_s3
    @patch('subprocess.check_output')
    def test_transform_s3_xml_simple(self, mocked_subprocess):
        """Test Pulling S3 XML, Transforming with XSLT, & Writing to S3."""
        # setup kwargs for test runs
        access_id = self.kwargs.get("access_id")
        access_secret = self.kwargs.get("access_secret")
        bucket = self.kwargs.get("bucket")
        test_key = self.kwargs.get("source_prefix") + "/xsl-sample.xml"

        # create expected mocked s3 resources
        conn = boto3.client("s3", aws_access_key_id=access_id, aws_secret_access_key=access_secret)
        conn.create_bucket(Bucket=bucket)
        conn.put_object(Bucket=bucket, Key=test_key, Body=open("tests/fixtures/xsl-sample.xml").read())
        test_content_exists = conn.get_object(Bucket=bucket, Key=test_key)
        test_object_exists = conn.list_objects(Bucket=bucket)
        self.assertEqual(test_content_exists["Body"].read(), open("tests/fixtures/xsl-sample.xml", "rb").read())
        self.assertEqual(test_content_exists["ResponseMetadata"]["HTTPStatusCode"], 200)
        self.assertEqual(test_object_exists["Contents"][0]["Key"], test_key)

        # setup mocked subprocess result
        mocked_subprocess.side_effect = [
            open("tests/fixtures/xsl-sample-simple-output-record1.xml", "rb").read(),
            open("tests/fixtures/xsl-sample-simple-output-record2.xml", "rb").read(),
            open("tests/fixtures/xsl-sample-simple-output-record3.xml", "rb").read()
        ]

        # run tests
        with self.assertLogs() as log:
            transform.transform_s3_xsl(**self.kwargs)
        self.assertIn("INFO:root:Transforming File dpla_test/new-updated-filtered/xsl-sample.xml", log.output)
        self.assertIn("INFO:root:Transforming Record oai:digital.library.villanova.edu:vudl:293113", log.output)
        self.assertIn("INFO:root:Transforming Record oai:digital.library.villanova.edu:vudl:469533", log.output)
        self.assertIn("INFO:root:Transforming Record oai:digital.library.villanova.edu:vudl:469545", log.output)
        test_output_objects = conn.list_objects(Bucket=bucket, Prefix=self.kwargs.get("destination_prefix"))
        self.assertEqual(test_output_objects["ResponseMetadata"]["HTTPStatusCode"], 200)
        test_output_objects_ar = [object.get("Key") for object in test_output_objects["Contents"]]
        self.assertEqual(test_output_objects_ar, ["dpla_test/transformed/xsl-sample.xml"])
        test_output_content = etree.fromstring(conn.get_object(Bucket=bucket, Key="dpla_test/transformed/xsl-sample.xml")["Body"].read())
        should_match_output = etree.fromstring(open("tests/fixtures/xsl-sample-simple-output-all.xml", "rb").read())
        self.assertEqual(
            etree.tostring(test_output_content, pretty_print=True),
            etree.tostring(should_match_output, pretty_print=True)
        )


    @mock_s3
    @patch('subprocess.check_output')
    def test_transform_s3_xml_complex(self, mocked_subprocess):
        """Test Pulling S3 XML, Transforming with Complex XSLT, & Writing to S3."""
        # setup kwargs for test runs
        access_id = self.kwargs.get("access_id")
        access_secret = self.kwargs.get("access_secret")
        bucket = self.kwargs.get("bucket")
        test_key = self.kwargs.get("source_prefix") + "/xsl-sample.xml"

        # create expected mocked s3 resources
        conn = boto3.client("s3", aws_access_key_id=access_id, aws_secret_access_key=access_secret)
        conn.create_bucket(Bucket=bucket)
        conn.put_object(Bucket=bucket, Key=test_key, Body=open("tests/fixtures/xsl-sample.xml").read())
        test_content_exists = conn.get_object(Bucket=bucket, Key=test_key)
        test_object_exists = conn.list_objects(Bucket=bucket)
        self.assertEqual(test_content_exists["Body"].read(), open("tests/fixtures/xsl-sample.xml", "rb").read())
        self.assertEqual(test_content_exists["ResponseMetadata"]["HTTPStatusCode"], 200)
        self.assertEqual(test_object_exists["Contents"][0]["Key"], test_key)

        # setup mocked subprocess result
        mocked_subprocess.side_effect = [
            open("tests/fixtures/xsl-sample-complex-output-record1.xml", "rb").read(),
            open("tests/fixtures/xsl-sample-complex-output-record2.xml", "rb").read(),
            open("tests/fixtures/xsl-sample-complex-output-record3.xml", "rb").read()
        ]

        # run tests
        with self.assertLogs() as log:
            transform.transform_s3_xsl(**self.kwargs)
        self.assertIn("INFO:root:Transforming File dpla_test/new-updated-filtered/xsl-sample.xml", log.output)
        self.assertIn("INFO:root:Transforming Record oai:digital.library.villanova.edu:vudl:293113", log.output)
        self.assertIn("INFO:root:Transforming Record oai:digital.library.villanova.edu:vudl:469533", log.output)
        self.assertIn("INFO:root:Transforming Record oai:digital.library.villanova.edu:vudl:469545", log.output)
        test_output_objects = conn.list_objects(Bucket=bucket, Prefix=self.kwargs.get("destination_prefix"))
        self.assertEqual(test_output_objects["ResponseMetadata"]["HTTPStatusCode"], 200)
        test_output_objects_ar = [object.get("Key") for object in test_output_objects["Contents"]]
        self.assertEqual(test_output_objects_ar, ["dpla_test/transformed/xsl-sample.xml"])
        test_output_content = etree.fromstring(conn.get_object(Bucket=bucket, Key="dpla_test/transformed/xsl-sample.xml")["Body"].read())
        should_match_output = etree.fromstring(open("tests/fixtures/xsl-sample-complex-output-all.xml", "rb").read())
        self.assertEqual(
            etree.tostring(test_output_content, pretty_print=True),
            etree.tostring(should_match_output, pretty_print=True)
        )



"""tulflow Python package setup."""

import os
import sys
import setuptools
from setuptools.command.install import install

VERSION = "v0.8.6"

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version"""
    description = 'verify that the git tag matches our version'

    def run(self):
        tag = os.getenv('CIRCLE_TAG')

        if tag != VERSION:
            info = "Git tag: {0} does not match the version of this app: {1}".format(
                tag, VERSION
            )
            sys.exit(info)

setuptools.setup(
    name="tulflow",
    author="Temple University Libraries",
    author_email="tul08567@temple.edu",
    description="Package of Temple University Library Indexing & ETL functions used by Airflow.",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/tulibraries/tulflow",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    test_suite="nose.collector",
    tests_require=["nose"],
    setup_requires=["setuptools>=17.1"],
	version=VERSION,
	cmdclass={
        "verify": VerifyVersionCommand,
    }
)



"""Tests suite for tulflow harvest (Functions for harvesting OAI in Airflow Tasks)."""
import json
from re import compile as re_compile
import unittest
import requests_mock
from tulflow.solr_api_utils import SolrApiUtils
from unittest.mock import Mock, patch


class TestSolrApiUtils(unittest.TestCase):
    """Test Class for Solr Cloud api functions."""
    def setUp(self):
        self.sc_url = "https://sc.example.com"
        self.matcher = re_compile(f"{self.sc_url}/.*")
        self.solrcloud = SolrApiUtils(self.sc_url)

    @requests_mock.mock()
    def test_get_configsets(self, rm):
        response = json.dumps({"configSets":["test", "_default", "test2"]})
        rm.get(self.matcher, text=response)
        configsets = self.solrcloud.get_configsets()
        self.assertEqual(configsets, ["test", "_default", "test2"])

    @requests_mock.mock()
    def test_get_collections(self, rm):
        response = json.dumps({
            "responseHeader": {"status":0, "QTime":1},
            "collections": ["test_collection1", "test_collection2"]
        })
        rm.get(self.matcher, text=response)
        collections = self.solrcloud.get_collections()
        self.assertEqual(collections, ["test_collection1", "test_collection2"])

    @requests_mock.mock()
    def test_get_aliases(self, rm):
        response = json.dumps({
            "responseHeader": {"status":0, "QTime":1},
            "aliases": {
                "test_alias1": "test_coll1,test_coll2",
                "test_alias2": "test_coll3,test_coll4"
            }
        })
        rm.get(self.matcher, text=response)
        aliases = self.solrcloud.get_aliases()
        self.assertEqual(aliases, {
            "test_alias1": "test_coll1,test_coll2",
            "test_alias2": "test_coll3,test_coll4"
        })

    @patch('tulflow.solr_api_utils.SolrApiUtils.get_configsets')
    def test_most_recent_configsets(self, mock_get_configsets):
        mock_get_configsets.return_value = ['tul_cob-catalog-9', 'tul_cob-catalog-10', 'tul_cob-web-2', '_default', 'tul_cob-web-3']
        most_recent = self.solrcloud.most_recent_configsets()
        self.assertEqual(most_recent, ["tul_cob-catalog-10", "tul_cob-web-3"])

    @patch('tulflow.solr_api_utils.SolrApiUtils.get_alias_collections')
    def test_is_collection_in_alias(self, mock_get_alias_collections):
        alias = 'funcake-oai'
        mock_get_alias_collections.return_value = ['funcake-villanova', 'funcake-temple']
        is_present = self.solrcloud.is_collection_in_alias('funcake-villanova', alias)
        self.assertTrue(is_present)

        is_not_present = self.solrcloud.is_collection_in_alias('funcake-yale', alias)
        self.assertFalse(is_not_present)

    @patch('tulflow.solr_api_utils.SolrApiUtils.get_collections')
    def test_collection_exists(self, mock_get_collections):
        mock_get_collections.return_value = ['coll1', 'coll2', 'coll3']
        being = self.solrcloud.collection_exists('coll3')
        self.assertTrue(being)
        nothingness = self.solrcloud.collection_exists('sartre')
        self.assertFalse(nothingness)

    def test_filter_init_collection(self):
        aliases = ["test-init", "test-real", "test-realer"]
        #uses default of ending with "-init"
        filtered = self.solrcloud.filter_init_collection(aliases)
        self.assertFalse("test-init" in filtered)
        # Also accepts specific name
        filtered = self.solrcloud.filter_init_collection(aliases, "test-realer")
        self.assertFalse("test-realer" in filtered)

    @patch('tulflow.solr_api_utils.SolrApiUtils.create_or_modify_alias_and_set_collections')
    @patch('tulflow.solr_api_utils.SolrApiUtils.get_alias_collections')
    def test_remove_collection_from_alias(self, mock_get_alias_collections, mock_create_or_modify_alias_and_set_collections):
        alias = "funcake-oai"
        mock_get_alias_collections.return_value = ['funcake-villanova', 'funcake-temple']
        self.solrcloud.remove_collection_from_alias('funcake-villanova', alias)
        mock_create_or_modify_alias_and_set_collections.assert_called_with(alias=alias, collections=['funcake-temple'])
        # make sure we don't error out of the collection is not part of the alias already
        try:
            self.solrcloud.remove_collection_from_alias('funcake-not-present', alias)
        except Exception as e:
            self.fail(f"Removing a collection not present in the alias raised an error: {e}")


    @patch('tulflow.solr_api_utils.SolrApiUtils.create_or_modify_alias_and_set_collections')
    @patch('tulflow.solr_api_utils.SolrApiUtils.get_alias_collections')
    def test_add_collection_to_alias(self, mock_get_alias_collections, mock_create_or_modify_alias_and_set_collections):
        alias = "funcake-oai"
        mock_get_alias_collections.return_value = ['funcake-temple']
        self.solrcloud.add_collection_to_alias('funcake-villanova', alias)
        mock_create_or_modify_alias_and_set_collections.assert_called_with(alias=alias, collections=['funcake-temple', 'funcake-villanova'])


    @patch('tulflow.solr_api_utils.SolrApiUtils.remove_collection_from_alias')
    @patch('tulflow.solr_api_utils.SolrApiUtils.delete_collection')
    @patch('tulflow.solr_api_utils.SolrApiUtils.create_collection')
    @patch('tulflow.solr_api_utils.SolrApiUtils.add_collection_to_alias')
    def test_remove_and_recreate_collection_from_alias(self, mock_add_collection_to_alias, mock_create_collection, mock_delete_collection, mock_remove_collection_from_alias):
        collection = "test_collection"
        alias = "test_alias"
        configset = "test_configset"
        SolrApiUtils.remove_and_recreate_collection_from_alias(collection=collection, configset=configset, alias=alias, solr_url="http://sc.example.com", solr_port="443", replicationFactor=3)
        mock_remove_collection_from_alias.assert_called_with(collection=collection, alias=alias)
        mock_delete_collection.assert_called_with(collection)
        mock_create_collection.assert_called_with(collection=collection, configset=configset, replicationFactor=3)
        mock_add_collection_to_alias.assert_called_with(collection=collection, alias=alias)

    @patch('tulflow.solr_api_utils.SolrApiUtils.get_from_solr_api')
    def test_create_collection(self, mock_get_from_solr_api):
        collection = "test_collection"
        configset = "test_configset"
        numShards = 5
        replicationFactor = 5
        maxShardsPerNode = 2
        self.solrcloud.create_collection(collection=collection, configset=configset, numShards=numShards, replicationFactor=replicationFactor, maxShardsPerNode=maxShardsPerNode)
        path = f"/solr/admin/collections?action=CREATE&name={collection}&collection.configName={configset}&numShards={numShards}&replicationFactor={replicationFactor}&maxShardsPerNode={maxShardsPerNode}"
        mock_get_from_solr_api.assert_called_with(path)
